from __future__ import annotations

import json
from typing import List, Optional

from pydantic import AnyHttpUrl, BaseModel, Field, SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict


class OIDCProvider(BaseModel):
    name: str
    issuer: str
    client_id: str
    client_secret: SecretStr
    scope: str = "openid email profile"


class JWTSettings(BaseModel):
    secret: SecretStr
    lifetime_seconds: int = 60 * 60 * 24 * 7


class PasswordClient(BaseModel):
    client_id: str
    client_secret: SecretStr


class AuthSettings(BaseSettings):
    # ---- JWT ----
    jwt: Optional[JWTSettings] = None

    # ---- Password login ----
    password_clients: List[PasswordClient] = Field(default_factory=list)
    require_client_secret_on_password_login: bool = False

    # ---- MFA / TOTP ----
    mfa_default_enabled_for_new_users: bool = False
    mfa_enforce_for_all_users: bool = False
    mfa_enforce_for_tenants: list[str] = []
    mfa_issuer: str = "svc-infra"
    mfa_pre_token_lifetime_seconds: int = 300
    mfa_recovery_codes: int = 8
    mfa_recovery_code_length: int = 10

    # ---- Email OTP ----
    email_otp_ttl_seconds: int = 5 * 60
    email_otp_cooldown_seconds: int = 60
    email_otp_attempts: int = 5

    # ---- Email/SMTP (verification, reset, etc.) ----
    smtp_host: Optional[str] = None
    smtp_port: int = 587
    smtp_username: Optional[str] = None
    smtp_password: Optional[SecretStr] = None
    smtp_from: Optional[str] = None

    # Dev convenience: auto-verify users without sending email
    auto_verify_in_dev: bool = True

    # ---- Built-in provider creds (optional) ----
    google_client_id: Optional[str] = None
    google_client_secret: Optional[SecretStr] = None
    github_client_id: Optional[str] = None
    github_client_secret: Optional[SecretStr] = None
    ms_client_id: Optional[str] = None
    ms_client_secret: Optional[SecretStr] = None
    ms_tenant: Optional[str] = None
    li_client_id: Optional[str] = None
    li_client_secret: Optional[SecretStr] = None
    oidc_providers: List[OIDCProvider] = Field(default_factory=list)

    # ---- Redirect + cookie settings ----
    post_login_redirect: AnyHttpUrl | str = "http://localhost:3000/app"
    redirect_allow_hosts_raw: str = "localhost,127.0.0.1"

    session_cookie_name: str = "svc_session"
    auth_cookie_name: str = "svc_auth"
    session_cookie_secure: bool = False
    session_cookie_samesite: str = "lax"
    session_cookie_domain: Optional[str] = None
    session_cookie_max_age_seconds: int = 60 * 60 * 4

    model_config = SettingsConfigDict(
        env_prefix="AUTH_",
        env_file=".env",
        extra="ignore",
        env_nested_delimiter="__",
    )


_settings: AuthSettings | None = None


def get_auth_settings() -> AuthSettings:
    global _settings
    if _settings is None:
        _settings = AuthSettings()
    return _settings


def parse_redirect_allow_hosts(raw: str | None) -> list[str]:
    if not raw:
        return ["localhost", "127.0.0.1"]
    s = raw.strip()
    if s.startswith("["):
        try:
            val = json.loads(s)
            if isinstance(val, list):
                return [str(x).strip() for x in val if str(x).strip()]
        except Exception:
            pass
    return [h.strip() for h in s.split(",") if h.strip()]



from __future__ import annotations

import re
from typing import Sequence

# Environment variable names to look up for DB URL
DEFAULT_DB_ENV_VARS: Sequence[str] = (
    "SQL_URL",
    "DB_URL",
)

# Regex used to detect async drivers from URL drivername
ASYNC_DRIVER_HINT = re.compile(r"\+(?:async|asyncpg|aiosqlite|aiomysql|asyncmy|aio\w+)")

# Alembic templates loaded from package resources (svc_infra.db.sql.templates.setup)
# Kept as module-level constants for compatibility with core.py
try:
    import importlib.resources as pkg

    _tmpl_pkg = pkg.files("svc_infra.db.sql.templates.setup")
    ALEMBIC_INI_TEMPLATE = _tmpl_pkg.joinpath("alembic.ini.tmpl").read_text(encoding="utf-8")
    ALEMBIC_SCRIPT_TEMPLATE = _tmpl_pkg.joinpath("script.py.mako.tmpl").read_text(encoding="utf-8")
except Exception:
    # Fallbacks (should not normally happen). Provide minimal safe defaults.
    ALEMBIC_INI_TEMPLATE = (
        """[alembic]\nscript_location = {script_location}\nsqlalchemy.url = {sqlalchemy_url}\n"""
    )
    ALEMBIC_INI_TEMPLATE = (
        """[alembic]\nscript_location = {script_location}\nsqlalchemy.url = {sqlalchemy_url}\n"""
    )
    ALEMBIC_SCRIPT_TEMPLATE = '"""${message}"""\nfrom alembic import op\nimport sqlalchemy as sa\n\nrevision = ${repr(up_revision)}\ndown_revision = ${repr(down_revision)}\nbranch_labels = ${repr(branch_labels)}\ndepends_on = ${repr(depends_on)}\n\ndef upgrade():\n    ${upgrades if upgrades else "pass"}\n\n\ndef downgrade():\n    ${downgrades if downgrades else "pass"}\n'
__all__ = [
    "DEFAULT_DB_ENV_VARS",
    "ASYNC_DRIVER_HINT",
    "ALEMBIC_INI_TEMPLATE",
    "ALEMBIC_SCRIPT_TEMPLATE",
]



from __future__ import annotations

from typing import Optional

from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase

from .settings import MongoSettings

_client: Optional[AsyncIOMotorClient] = None
_db: Optional[AsyncIOMotorDatabase] = None


def _client_opts(cfg: MongoSettings) -> dict:
    return {
        "appname": cfg.appname,
        "minPoolSize": cfg.min_pool_size,
        "maxPoolSize": cfg.max_pool_size,
        "uuidRepresentation": "standard",
    }


async def init_mongo(cfg: MongoSettings | None = None) -> AsyncIOMotorDatabase:
    global _client, _db
    cfg = cfg or MongoSettings()
    if _client is None:
        if not cfg.db_name:
            raise RuntimeError("MONGO_DB must be set.")
        _client = AsyncIOMotorClient(str(cfg.url), **_client_opts(cfg))
        _db = _client[cfg.db_name]
    assert _db is not None
    return _db


async def acquire_db() -> AsyncIOMotorDatabase:
    global _db
    if _db is None:
        await init_mongo()
    assert _db is not None
    return _db


async def ping_mongo() -> bool:
    db = await acquire_db()
    res = await db.command("ping")
    return bool(res and res.get("ok") == 1)


async def close_mongo() -> None:
    global _client, _db
    if _client is not None:
        _client.close()
    _client = None
    _db = None



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, List, Optional

FAVICON_DATA_URI = (
    "data:image/svg+xml,"
    "%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 64 64'%3E"
    "%3Cpath fill='%230b0f14' d='M0 0h64v64H0z'/%3E"
    "%3Cpath fill='%23e6edf3' d='M21 16c-6 4-9 8-9 16s3 12 9 16l3-4c-4-3-6-6-6-12s2-9 6-12l-3-4zm22 0-3 4c4 3 6 6 6 12s-2 9-6 12l3 4c6-4 9-8 9-16s-3-12-9-16z'/%3E"
    "%3C/svg%3E"
)


@dataclass(frozen=True)
class DocTargets:
    swagger: Optional[str] = None
    redoc: Optional[str] = None
    openapi_json: Optional[str] = None


@dataclass(frozen=True)
class CardSpec:
    tag: str
    docs: DocTargets


def _btn(label: str, href: str) -> str:
    return f'<a class="btn" href="{href}">{label}</a>'


def _card(spec: CardSpec) -> str:
    tag = "/" if spec.tag.strip("/") == "" else f"/{spec.tag.strip('/')}"
    links: List[str] = []
    if spec.docs.swagger:
        links.append(_btn("Swagger", spec.docs.swagger))
    if spec.docs.redoc:
        links.append(_btn("ReDoc", spec.docs.redoc))
    if spec.docs.openapi_json:
        links.append(_btn("OpenAPI JSON", spec.docs.openapi_json))
    actions = "\n".join(links) if links else "<div class='muted'>No docs exposed</div>"

    return f"""
    <div class="card">
      <div class="card__body">
        <div class="chip">{tag}</div>
        <div class="actions">{actions}</div>
      </div>
    </div>
    """.strip()


def render_index_html(*, service_name: str, release: str, cards: Iterable[CardSpec]) -> str:
    grid = "\n".join(_card(c) for c in cards)
    return f"""
<!doctype html>
<html>
<head>
  <meta charset="utf-8"/>
  <meta name="viewport" content="width=device-width, initial-scale=1"/>
  <title>{service_name} • {release}</title>
  <link rel="icon" href="{FAVICON_DATA_URI}">
  <style>
    :root {{
      --bg:#0b0f14; --fg:#e6edf3; --muted:#9aa7b3; --panel:#0f141b; --panel-2:#0d1218;
      --border:#1f2631; --border-strong:#2b3546; --btn-bg:#121a24; --btn-fg:#d7e2ee;
      --btn-border:#223044; --btn-hover-bg:#162130; --btn-hover-border:#2a3b54; --radius:12px;
    }}
    @media (prefers-color-scheme: light) {{
      :root {{
        --bg:#f7f9fc; --fg:#0b0f14; --muted:#556171; --panel:#ffffff; --panel-2:#fafbfc;
        --border:#e6ebf2; --border-strong:#d7dfeb; --btn-bg:#f6f8fc; --btn-fg:#0b1220;
        --btn-border:#dce5f2; --btn-hover-bg:#eef3fb; --btn-hover-border:#cfdbee;
      }}
    }}
    * {{ box-sizing: border-box; }}
    body {{
      margin: 0; padding: 28px; background: var(--bg); color: var(--fg);
      font: 500 15px/1.5 system-ui, -apple-system, Segoe UI, Inter, Roboto, Helvetica, Arial;
    }}
    .container {{ max-width: 1100px; margin: 0 auto; }}
    .header {{ display:flex; align-items:baseline; gap:12px; margin-bottom:18px; }}
    .title {{ margin:0; font-size:24px; letter-spacing:.2px; }}
    .sub {{ color:var(--muted); font-size:14px; }}
    .grid {{ display:grid; grid-template-columns:repeat(auto-fill, minmax(240px,1fr)); gap:14px; }}
    .card {{
      background: linear-gradient(180deg, var(--panel), var(--panel-2));
      border: 1px solid var(--border); border-radius: var(--radius);
      transition: border-color .12s ease, transform .08s ease;
    }}
    .card:hover {{ border-color: var(--border-strong); transform: translateY(-1px); }}
    .card__body {{ padding: 16px; display:grid; gap:12px; }}
    .chip {{
      display:inline-block; font-size:15px; font-weight:600; color:var(--fg);
      background:#1c2736; border:1px solid var(--border-strong);
      border-radius: 999px; padding:6px 14px; width:fit-content; letter-spacing:.3px;
    }}
    .actions {{ display:flex; gap:8px; flex-wrap:wrap; align-items:center; }}
    .btn {{
      display:inline-block; text-decoration:none; font-weight:500; border-radius:8px;
      padding:6px 10px; font-size:13px; border:1px solid var(--btn-border);
      background:var(--btn-bg); color:var(--btn-fg);
      transition: transform .06s ease, border-color .12s ease, background .12s ease;
    }}
    .btn:hover {{ transform: translateY(-1px); background: var(--btn-hover-bg); border-color: var(--btn-hover-border); }}
    footer {{ margin-top:22px; color:var(--muted); font-size:13px; }}
  </style>
</head>
<body>
  <div class="container">
    <header class="header">
      <h1 class="title">{service_name}</h1>
      <div class="sub">release {release}</div>
    </header>
    <section class="grid">
      {grid}
    </section>
    <footer>Tip: each card exposes Swagger, ReDoc, and a pretty JSON view.</footer>
  </div>
</body>
</html>
    """.strip()



"""
Cache module providing decorators and utilities for caching operations.

This module offers high-level decorators for read/write caching, cache invalidation,
and resource-based cache management.
"""

# Core decorators - main public API
from .decorators import cached  # alias for cache_read
from .decorators import mutates  # alias for cache_write
from .decorators import cache_read, cache_write, init_cache, init_cache_async

# Recaching functionality for advanced use cases
from .recache import RecachePlan, recache

# Resource management for entity-based caching
from .resources import entity  # legacy alias
from .resources import resource

__all__ = [
    # Primary decorators developers use
    "cache_read",
    "cached",
    "cache_write",
    "mutates",
    # Cache initialization
    "init_cache",
    "init_cache_async",
    # Advanced recaching
    "RecachePlan",
    "recache",
    # Resource-based caching
    "resource",
    "entity",
]



from __future__ import annotations

from typing import Any, Awaitable, Callable, Optional, Sequence

import jwt
from fastapi import Depends, HTTPException, Request

from ..auth.settings import get_auth_settings

AUDIENCE = ["fastapi-users:auth"]

GetSessionDep = Any  # FastAPI dep marker provided by adapters
GetUserById = Callable[[Any, Any], Awaitable[Optional[Any]]]
GetUserRoles = Callable[[Any], Sequence[str]]


def make_optional_user_dep(*, session_dep: GetSessionDep, get_user_by_id: GetUserById):
    async def optional_user(request: Request, session=Depends(session_dep)):
        st = get_auth_settings()
        cookie_name = getattr(st, "auth_cookie_name", "svc_auth")
        raw = request.cookies.get(cookie_name)
        if not raw:
            request.state.user = None
            return None
        try:
            payload = jwt.decode(
                raw, st.jwt.secret.get_secret_value(), algorithms=["HS256"], audience=AUDIENCE
            )
        except Exception:
            request.state.user = None
            return None
        user = await get_user_by_id(session, payload.get("sub"))
        request.state.user = user
        return user

    return optional_user


def make_require_user_dep(optional_user_dep):
    async def require_user(user=Depends(optional_user_dep)):
        if not user:
            raise HTTPException(401, "unauthenticated")
        return user

    return require_user


def make_require_roles_dep(*, require_user_dep, get_user_roles: GetUserRoles, roles: Sequence[str]):
    async def require_roles(user=Depends(require_user_dep)):
        if not set(roles).issubset(set(get_user_roles(user) or [])):
            raise HTTPException(403, "forbidden")
        return user

    return require_roles



from __future__ import annotations

import logging
import os
from logging.config import dictConfig
from typing import Sequence

from svc_infra.app.env import CURRENT_ENVIRONMENT

from .filter import filter_logs_for_paths
from .formats import (
    JsonFormatter,
    LoggingConfig,
    _env_name_list_to_enum_values,
    _parse_paths_csv,
    _read_format,
    _read_level,
)


def setup_logging(
    level: str | None = None,
    fmt: str | None = None,
    *,
    drop_paths: Sequence[str] | None = None,
    filter_envs: Sequence[str] | None = ("prod", "test"),
) -> None:
    """Configure logging + optional access-log path filtering."""
    if fmt is not None or level is not None:
        LoggingConfig(fmt=fmt, level=level)  # pydantic validation

    if level is None:
        level = _read_level()
    if fmt is None:
        fmt = _read_format()

    formatter_name = "json" if fmt == "json" else "plain"

    # Silence chattier libs outside DEBUG
    if level.upper() != "DEBUG":
        logging.getLogger("multipart.multipart").setLevel(logging.WARNING)

    dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "plain": {
                    "format": "%(asctime)s %(levelname)-5s [pid:%(process)d] %(name)s: %(message)s",
                    "datefmt": "%Y-%m-%dT%H:%M:%S",
                },
                "json": {
                    "()": JsonFormatter,
                    "datefmt": "%Y-%m-%dT%H:%M:%S",
                },
            },
            "handlers": {
                "stream": {
                    "class": "logging.StreamHandler",
                    "level": level,
                    "formatter": formatter_name,
                }
            },
            "root": {"level": level, "handlers": ["stream"]},
            "loggers": {
                "uvicorn": {"level": "INFO", "handlers": [], "propagate": True},
                "uvicorn.error": {"level": "INFO", "handlers": [], "propagate": True},
                "uvicorn.access": {"level": "INFO", "handlers": [], "propagate": True},
            },
        }
    )

    # Access-log path filter
    enabled_envs = _env_name_list_to_enum_values(filter_envs)
    filter_enabled = CURRENT_ENVIRONMENT.value in enabled_envs

    env_paths = _parse_paths_csv(os.getenv("LOG_DROP_PATHS"))
    if drop_paths is not None:
        paths = list(drop_paths)
    elif env_paths:
        paths = env_paths
    else:
        paths = ["/metrics"] if filter_enabled else []

    filter_logs_for_paths(paths=paths, enabled=filter_enabled)



from __future__ import annotations

import pyotp
from fastapi import APIRouter, Body, Depends, HTTPException, Request, Security
from pydantic import BaseModel
from starlette.responses import JSONResponse

from svc_infra.api.fastapi import public_router
from svc_infra.api.fastapi.auth.pre_auth import get_mfa_pre_jwt_writer
from svc_infra.api.fastapi.auth.settings import get_auth_settings
from svc_infra.api.fastapi.db.sql.session import SqlSessionDep

from .mfa import EMAIL_OTP_STORE, _hash, _now_utc_ts  # reuse helpers
from .security import cookie_auth_optional, oauth2_scheme_optional


class RequireMFAIn(BaseModel):
    code: str | None = None
    pre_token: str | None = None


class DisableAccountIn(RequireMFAIn):
    reason: str | None = None


class DeleteAccountIn(RequireMFAIn):
    hard: bool = False  # soft by default


def account_router(
    *,
    user_model: type,
    auth_prefix: str = "/auth",
) -> APIRouter:
    router = public_router(prefix=f"{auth_prefix}/account", tags=["auth:account"])

    async def _current_user(request: Request, session: SqlSessionDep):
        st = get_auth_settings()
        token = request.headers.get("authorization", "").removeprefix(
            "Bearer "
        ).strip() or request.cookies.get(st.auth_cookie_name)
        if not token:
            raise HTTPException(401, "Missing token")

        # Read token via your strategy (same as elsewhere)
        from svc_infra.api.fastapi.db.sql.users import get_fastapi_users

        fapi, auth_backend, *_ = get_fastapi_users(
            user_model, None, None, None, public_auth_prefix=auth_prefix
        )
        # IMPORTANT: get a real user_manager instance to pass into read_token
        user_manager_gen = fapi.get_user_manager
        async for user_manager in user_manager_gen():  # type: ignore
            strategy = auth_backend.get_strategy()
            try:
                user = await strategy.read_token(token, user_manager)
            finally:
                break  # we only need one instance

        if not user:
            raise HTTPException(401, "Invalid token")

        db_user = await session.get(user_model, user.id)
        if not db_user:
            raise HTTPException(401, "Invalid token")
        if not getattr(db_user, "is_active", True):
            raise HTTPException(401, "account_disabled")

        return db_user, session

    async def _verify_code_for_user(user, code: str | None, pre_token: str | None) -> bool:
        """Accept TOTP, recovery, or email OTP (requires pre_token for email)."""
        if not code:
            return False
        # TOTP
        if getattr(user, "mfa_secret", None):
            if pyotp.TOTP(user.mfa_secret).verify(code, valid_window=1):
                return True
        # Recovery
        dig = _hash(code)
        recov = getattr(user, "mfa_recovery", None) or []
        if dig in recov:
            recov.remove(dig)
            return True
        # Email OTP
        if pre_token:
            try:
                pre = await get_mfa_pre_jwt_writer().read(pre_token)
                uid = str(pre.get("sub") or "")
            except Exception:
                uid = ""
            if uid and uid == str(user.id):
                rec = EMAIL_OTP_STORE.get(uid)
                now = _now_utc_ts()
                if (
                    rec
                    and now <= rec["exp"]
                    and rec["attempts_left"] > 0
                    and _hash(code) == rec["hash"]
                ):
                    EMAIL_OTP_STORE.pop(uid, None)
                    return True
                if rec:
                    rec["attempts_left"] = max(0, rec["attempts_left"] - 1)
        return False

    @router.post(
        "/disable",
        openapi_extra={
            "security": [{"OAuth2PasswordBearer": []}, {"cookieAuth": []}, {"APIKeyHeader": []}]
        },
    )
    async def disable_account(
        _b: str | None = Security(oauth2_scheme_optional),
        _c: str | None = Security(cookie_auth_optional),
        payload: DisableAccountIn | None = Body(None),
        dep=Depends(_current_user),
    ):
        user, session = dep

        if getattr(user, "mfa_enabled", False):
            if not payload or not await _verify_code_for_user(
                user, payload.code, payload.pre_token
            ):
                raise HTTPException(400, "Invalid code")
            await session.flush()  # burns recovery code if used

        reason = (payload.reason if payload else None) or "user_disabled_self"
        user.is_active = False
        user.disabled_reason = reason
        await session.commit()
        return JSONResponse({"ok": True})

    @router.post(
        "/delete",
        openapi_extra={
            "security": [{"OAuth2PasswordBearer": []}, {"cookieAuth": []}, {"APIKeyHeader": []}]
        },
    )
    async def delete_account(
        payload: DeleteAccountIn | None = Body(None), dep=Depends(_current_user)
    ):
        user, session = dep

        if getattr(user, "mfa_enabled", False):
            if not payload or not await _verify_code_for_user(
                user, payload.code, payload.pre_token
            ):
                raise HTTPException(400, "Invalid code")
            await session.flush()

        hard = bool(payload.hard) if payload else False
        if hard:
            await session.delete(user)
            await session.commit()
            return JSONResponse({"ok": True, "deleted": "hard"})
        else:
            user.is_active = False
            user.disabled_reason = "user_soft_deleted"
            await session.commit()
            return JSONResponse({"ok": True, "deleted": "soft"})

    return router



from __future__ import annotations

from typing import Any, Callable, Optional, Sequence

from fastapi import Depends

from ..dual_router import DualAPIRouter
from .deps import make_optional_user_dep, make_require_roles_dep, make_require_user_dep


def _merge(base: Optional[Sequence[Any]], extra: Optional[Sequence[Any]]) -> list[Any]:
    out: list[Any] = []
    if base:
        out.extend(base)
    if extra:
        out.extend(extra)
    return out


def optional_user(
    *,
    session_dep,
    get_user_by_id: Callable,
    dependencies: Optional[Sequence[Any]] = None,
    **kwargs: Any,
) -> DualAPIRouter:
    dep = make_optional_user_dep(session_dep=session_dep, get_user_by_id=get_user_by_id)
    return DualAPIRouter(dependencies=_merge([Depends(dep)], dependencies), **kwargs)


def require_user(
    *,
    session_dep,
    get_user_by_id: Callable,
    dependencies: Optional[Sequence[Any]] = None,
    **kwargs: Any,
) -> DualAPIRouter:
    opt = make_optional_user_dep(session_dep=session_dep, get_user_by_id=get_user_by_id)
    req = make_require_user_dep(opt)
    return DualAPIRouter(dependencies=_merge([Depends(req)], dependencies), **kwargs)


def require_roles(
    *roles: str,
    session_dep,
    get_user_by_id: Callable,
    get_user_roles: Callable,
    dependencies: Optional[Sequence[Any]] = None,
    **kwargs: Any,
) -> DualAPIRouter:
    opt = make_optional_user_dep(session_dep=session_dep, get_user_by_id=get_user_by_id)
    req = make_require_user_dep(opt)
    rol = make_require_roles_dep(require_user_dep=req, get_user_roles=get_user_roles, roles=roles)
    return DualAPIRouter(dependencies=_merge([Depends(rol)], dependencies), **kwargs)



from __future__ import annotations

from pathlib import Path
from typing import Optional, cast

import click
import typer

from svc_infra.db.sql.scaffold import (
    Kind,
    scaffold_core,
    scaffold_models_core,
    scaffold_schemas_core,
)


def cmd_scaffold(
    kind: str = typer.Option(
        "entity",
        "--kind",
        click_type=click.Choice(["entity", "auth"], case_sensitive=False),
    ),
    entity_name: str = typer.Option(
        "Item", help="Class name for entity/auth (e.g., User, Member, Product)."
    ),
    table_name: Optional[str] = typer.Option(
        None, help="Optional table name (defaults to plural snake of entity name)."
    ),
    models_dir: Path = typer.Option(..., help="Directory for models."),
    schemas_dir: Path = typer.Option(..., help="Directory for schemas."),
    overwrite: bool = typer.Option(False, help="Overwrite existing files."),
    same_dir: bool = typer.Option(
        False,
        "--same-dir/--no-same-dir",
        help="Put models & schemas into the same dir.",
    ),
    models_filename: Optional[str] = typer.Option(
        None, help="Custom filename for models (separate-dir mode)."
    ),
    schemas_filename: Optional[str] = typer.Option(
        None, help="Custom filename for schemas (separate-dir mode)."
    ),
):
    """
    Scaffold starter models/schemas for either:
    - kind=auth   → app/auth/models.py + schemas.py
    - kind=entity → app/models/<file>.py + app/schemas/<file>.py
    """
    res = scaffold_core(
        models_dir=models_dir,
        schemas_dir=schemas_dir,
        kind=cast(Kind, kind.lower()),
        entity_name=entity_name,
        table_name=table_name,
        overwrite=overwrite,
        same_dir=same_dir,
        models_filename=models_filename,
        schemas_filename=schemas_filename,
    )
    typer.echo(res)


def cmd_scaffold_models(
    dest_dir: Path = typer.Option(..., "--dest-dir", resolve_path=True),
    kind: str = typer.Option(
        "entity",
        "--kind",
        help="Scaffold type",
        click_type=click.Choice(["entity", "auth"], case_sensitive=False),
    ),
    entity_name: str = typer.Option("Item", "--entity-name"),
    table_name: Optional[str] = typer.Option(None, "--table-name"),
    include_tenant: bool = typer.Option(True, "--include-tenant/--no-include-tenant"),
    include_soft_delete: bool = typer.Option(
        False, "--include-soft-delete/--no-include-soft-delete"
    ),
    overwrite: bool = typer.Option(False, "--overwrite/--no-overwrite"),
    models_filename: Optional[str] = typer.Option(
        None,
        "--models-filename",
        help="Filename to write (e.g. project_models.py). Defaults to <snake(entity)>.py",
    ),
):
    """
    Scaffold starter SQLAlchemy models for either:
    - kind=auth   → app/auth/models.py
    - kind=entity → app/models/<file>.py
    """
    res = scaffold_models_core(
        dest_dir=dest_dir,
        kind=cast(Kind, kind.lower()),
        entity_name=entity_name,
        table_name=table_name,
        include_tenant=include_tenant,
        include_soft_delete=include_soft_delete,
        overwrite=overwrite,
        models_filename=models_filename,
    )
    typer.echo(res)


def cmd_scaffold_schemas(
    dest_dir: Path = typer.Option(..., "--dest-dir", resolve_path=True),
    kind: str = typer.Option(
        "entity",
        "--kind",
        help="Scaffold type",
        click_type=click.Choice(["entity", "auth"], case_sensitive=False),
    ),
    entity_name: str = typer.Option("Item", "--entity-name"),
    include_tenant: bool = typer.Option(True, "--include-tenant/--no-include-tenant"),
    overwrite: bool = typer.Option(False, "--overwrite/--no-overwrite"),
    schemas_filename: Optional[str] = typer.Option(
        None,
        "--schemas-filename",
        help="Filename to write (e.g. project_schemas.py). Defaults to <snake(entity)>.py",
    ),
):
    """
    Scaffold starter Pydantic schemas for either:
    - kind=auth   → app/auth/schemas.py
    - kind=entity → app/schemas/<file>.py
    """
    res = scaffold_schemas_core(
        dest_dir=dest_dir,
        kind=cast(Kind, kind.lower()),
        entity_name=entity_name,
        include_tenant=include_tenant,
        overwrite=overwrite,
        schemas_filename=schemas_filename,
    )
    typer.echo(res)


def register(app: typer.Typer) -> None:
    app.command("sql-scaffold")(cmd_scaffold)
    app.command("sql-scaffold-models")(cmd_scaffold_models)
    app.command("sql-scaffold-schemas")(cmd_scaffold_schemas)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Sequence

from motor.motor_asyncio import AsyncIOMotorDatabase
from pymongo import IndexModel

from svc_infra.db.nosql.indexes import normalize_indexes
from svc_infra.db.nosql.mongo.client import acquire_db, close_mongo, init_mongo, ping_mongo
from svc_infra.db.nosql.resource import NoSqlResource
from svc_infra.db.nosql.utils import (
    get_mongo_dbname_from_env,
    get_mongo_url_from_env,
    prepare_process_env,
)


async def _ping(db: AsyncIOMotorDatabase) -> None:
    res = await db.command("ping")
    if not res or res.get("ok") != 1:
        raise RuntimeError("Mongo ping failed")


async def _ensure_collections(db: AsyncIOMotorDatabase, names: Iterable[str]) -> None:
    existing = set(await db.list_collection_names())
    for name in names:
        if name not in existing:
            await db.create_collection(name)


async def _apply_indexes(
    db: AsyncIOMotorDatabase, *, collection: str, indexes: Sequence[IndexModel] | None
) -> list[str]:
    if not indexes:
        return []
    return await db[collection].create_indexes(list(indexes))


# collection + doc used to "lock" the chosen DB name for this app
_REG_DB = "__infra_registry__"
_REG_COLL = "db_locks"


async def assert_db_locked(
    db, expected_db_name: str, *, service_id: str, allow_rebind: bool = False
):
    registry = db.client.get_database(_REG_DB)
    await registry[_REG_COLL].create_index("service_id", unique=True)

    doc = await registry[_REG_COLL].find_one({"service_id": service_id}, projection={"db_name": 1})
    if doc is None:
        await registry[_REG_COLL].insert_one(
            {"service_id": service_id, "db_name": expected_db_name}
        )
        return

    locked = doc.get("db_name")
    if locked != expected_db_name and not allow_rebind:
        raise RuntimeError(
            f"Service '{service_id}' locked to Mongo DB '{locked}', but current target is '{expected_db_name}'. "
            f"Use allow_rebind=True if you intend to move."
        )
    if allow_rebind and locked != expected_db_name:
        await registry[_REG_COLL].update_one(
            {"service_id": service_id},
            {"$set": {"db_name": expected_db_name}},
        )


@dataclass(frozen=True)
class PrepareResult:
    ok: bool
    created_collections: list[str]
    created_indexes: dict[str, list[str]]


async def prepare_mongo(
    *,
    resources: Sequence[NoSqlResource],
    service_id: str,
    allow_rebind: bool = False,
) -> PrepareResult:
    """
    Ensure Mongo is reachable, collections exist, and indexes are applied.
    Resources carry their own index definitions (IndexModel or dict alias).
    """
    db = await acquire_db()
    if not await ping_mongo():
        raise RuntimeError("Mongo ping failed")

    expected_db = get_mongo_dbname_from_env(required=True)
    if db.name != expected_db:
        raise RuntimeError(f"Connected to Mongo DB '{db.name}', but env says '{expected_db}'.")

    await assert_db_locked(db, expected_db, service_id=service_id, allow_rebind=allow_rebind)

    # collections
    colls = [r.resolved_collection() for r in resources]
    await _ensure_collections(db, colls)
    created_colls = colls

    # indexes (from resource.indexes)
    created_idx: dict[str, list[str]] = {}
    for r in resources:
        coll = r.resolved_collection()
        idx_models = normalize_indexes(r.indexes)
        if idx_models:
            names = await _apply_indexes(db, collection=coll, indexes=idx_models)
            created_idx[coll] = names

    return PrepareResult(ok=True, created_collections=created_colls, created_indexes=created_idx)


def setup_and_prepare(
    *,
    resources: Sequence[NoSqlResource],
    service_id: str,
    allow_rebind: bool = False,
) -> dict:
    """Sync entrypoint: resolve env, init client, ensure collections & indexes, close client."""
    root = prepare_process_env(".")
    get_mongo_url_from_env(required=True)
    get_mongo_dbname_from_env(required=True)

    async def _run():
        await init_mongo()
        try:
            result = await prepare_mongo(
                resources=resources,
                service_id=service_id,
                allow_rebind=allow_rebind,
            )
            return {
                "ok": result.ok,
                "project_root": str(root),
                "created_collections": result.created_collections,
                "created_indexes": result.created_indexes,
            }
        finally:
            await close_mongo()

    import asyncio

    return asyncio.run(_run())



from __future__ import annotations

from typing import Literal, cast

from fastapi import FastAPI
from starlette.middleware.sessions import SessionMiddleware

from svc_infra.api.fastapi.auth.mfa import mfa_router
from svc_infra.api.fastapi.db.sql.users import get_fastapi_users
from svc_infra.app.env import CURRENT_ENVIRONMENT, DEV_ENV, LOCAL_ENV

from .. import Require
from .account import account_router
from .apikey_router import apikey_router
from .gaurd import login_client_guard, mfa_login_router
from .oauth_router import oauth_router_with_backend
from .policy import AuthPolicy, DefaultAuthPolicy
from .pre_auth import get_mfa_pre_jwt_writer
from .providers import providers_from_settings
from .settings import get_auth_settings


def add_auth(
    app: FastAPI,
    *,
    user_model,
    schema_read,
    schema_create,
    schema_update,
    post_login_redirect: str | None = None,
    auth_prefix: str = "/auth",
    oauth_prefix: str = "/auth/oauth",
    enable_password: bool = True,
    enable_oauth: bool = True,
    provider_account_model=None,
    auth_policy: AuthPolicy | None = None,
) -> None:
    (
        fapi,
        auth_backend,
        auth_router,
        users_router,
        get_strategy,
        register_router,
        verify_router,
        reset_router,
    ) = get_fastapi_users(
        user_model=user_model,
        user_schema_read=schema_read,
        user_schema_create=schema_create,
        user_schema_update=schema_update,
        public_auth_prefix=auth_prefix,
    )

    settings_obj = get_auth_settings()
    policy = auth_policy or DefaultAuthPolicy(settings_obj)
    include_in_docs = CURRENT_ENVIRONMENT in (LOCAL_ENV, DEV_ENV)

    if not any(m.cls.__name__ == "SessionMiddleware" for m in app.user_middleware):
        jwt_block = getattr(settings_obj, "jwt", None)
        secret = (
            jwt_block.secret.get_secret_value()
            if jwt_block and getattr(jwt_block, "secret", None)
            else "svc-dev-secret-change-me"
        )
        same_site_lit = cast(
            Literal["lax", "strict", "none"],
            str(getattr(settings_obj, "session_cookie_samesite", "lax")).lower(),
        )
        app.add_middleware(
            SessionMiddleware,
            secret_key=secret,
            session_cookie=getattr(settings_obj, "session_cookie_name", "svc_session"),
            max_age=getattr(settings_obj, "session_cookie_max_age_seconds", 4 * 3600),
            same_site=same_site_lit,
            https_only=bool(getattr(settings_obj, "session_cookie_secure", False)),
        )

    if enable_password:
        # MFA-aware password login (returns MFA_REQUIRED + pre_token when needed)
        app.include_router(
            mfa_login_router(
                fapi=fapi,
                auth_backend=auth_backend,
                user_model=user_model,
                get_mfa_pre_writer=get_mfa_pre_jwt_writer,
                public_auth_prefix=auth_prefix,
                auth_policy=policy,
            ),
            include_in_schema=include_in_docs,
            dependencies=[Require(login_client_guard)],
        )

        # Users + registration + email verify + password reset
        app.include_router(
            users_router,
            prefix=auth_prefix,
            tags=["users"],
            include_in_schema=include_in_docs,
        )
        app.include_router(
            register_router,
            prefix=auth_prefix,
            tags=["auth"],
            include_in_schema=include_in_docs,
        )
        app.include_router(
            verify_router,
            prefix=auth_prefix,
            tags=["auth"],
            include_in_schema=include_in_docs,
        )
        app.include_router(
            reset_router,
            prefix=auth_prefix,
            tags=["auth"],
            include_in_schema=include_in_docs,
        )

        # TOTP MFA endpoints (setup/confirm/disable/verify).
        # Mounting these doesn't force MFA; enforcement is driven by per-user flags
        # and any project/tenant-wide settings you apply in your login/OAuth flows.
        app.include_router(
            mfa_router(
                user_model=user_model,
                get_strategy=get_strategy,
                fapi=fapi,
                auth_prefix=auth_prefix,
            ),
            include_in_schema=include_in_docs,
        )

        # Account management (disable/delete)
        app.include_router(
            account_router(user_model=user_model, auth_prefix=auth_prefix),
            include_in_schema=include_in_docs,
        )
        app.include_router(apikey_router(), include_in_schema=include_in_docs)

    if enable_oauth:
        providers = providers_from_settings(settings_obj)
        if providers:
            app.include_router(
                oauth_router_with_backend(
                    user_model=user_model,
                    auth_backend=auth_backend,
                    providers=providers,
                    post_login_redirect=post_login_redirect
                    or getattr(settings_obj, "post_login_redirect", "/"),
                    prefix=oauth_prefix,
                    provider_account_model=provider_account_model,
                    auth_policy=policy,
                ),
                include_in_schema=include_in_docs,
            )



from datetime import datetime, timezone

from svc_infra.api.fastapi.auth.settings import get_auth_settings


def get_mfa_pre_jwt_writer():
    st = get_auth_settings()
    jwt_block = getattr(st, "jwt", None)

    # Force to plain string
    secret = (
        jwt_block.secret.get_secret_value()
        if jwt_block and getattr(jwt_block, "secret", None)
        else "svc-dev-secret-change-me"
    )
    secret = str(secret)

    lifetime = int(getattr(st, "mfa_pre_token_lifetime_seconds", 300))

    class PreTokenWriter:
        def __init__(self, secret: str, lifetime: int):
            self.secret = secret
            self.lifetime = lifetime

        async def write(self, user):
            from fastapi_users.jwt import generate_jwt

            now = datetime.now(timezone.utc)
            payload = {
                "sub": str(getattr(user, "id")),
                "aud": ["fastapi-users:mfa"],
                "iat": int(now.timestamp()),
                "exp": int(now.timestamp()) + self.lifetime,
            }
            return generate_jwt(payload, self.secret, algorithm="HS256")

        async def read(self, token: str):
            from fastapi_users.jwt import decode_jwt

            # IMPORTANT: pass a STRING, not a list
            return decode_jwt(token, self.secret, audience=["fastapi-users:mfa"])

    return PreTokenWriter(secret, lifetime)



from __future__ import annotations

import logging

from fastapi import Response, status

from svc_infra.api.fastapi import public_router

router = public_router()
ROUTER_TAG = "health"


@router.get("/ping", status_code=status.HTTP_200_OK)
def ping():
    logging.info("Health check: /ping endpoint accessed. Service is responsive.")
    return Response(status_code=status.HTTP_200_OK)



from __future__ import annotations

import os
from typing import Iterable, Sequence

from fastapi import FastAPI
from pydantic import BaseModel, Field

from svc_infra.app.env import pick
from svc_infra.app.logging import LogLevelOptions, setup_logging
from svc_infra.obs import add_observability

from .models import APIVersionSpec, ServiceInfo
from .setup import setup_service_api

# ---------- helpers ----------


def _env_bool(name: str, default: bool) -> bool:
    v = os.getenv(name)
    if v is None:
        return default
    return v.strip().lower() in ("1", "true", "yes", "on")


def _env_str(name: str) -> str | None:
    v = os.getenv(name)
    return v.strip() if v and v.strip() else None


def _env_csv_paths(name: str) -> list[str] | None:
    v = os.getenv(name)
    if not v:
        return None
    parts = []
    for p in v.replace(",", " ").split():
        p = p.strip()
        if not p:
            continue
        parts.append(p if p.startswith("/") else f"/{p}")
    return parts


# ---------- Options models ----------


class LoggingOptions(BaseModel):
    enable: bool = Field(default=True)
    # None -> auto (json in prod, plain elsewhere); level auto if None
    level: str | None = None
    fmt: str | None = None  # "json" | "plain" | None


class ObservabilityOptions(BaseModel):
    enable: bool = Field(default=True)
    # Optional extras (only used if enable=True)
    db_engines: Iterable[object] | None = None
    metrics_path: str | None = None
    skip_metric_paths: Iterable[str] | None = None


class EasyAppOptions(BaseModel):
    logging: LoggingOptions = LoggingOptions()
    observability: ObservabilityOptions = ObservabilityOptions()

    @classmethod
    def from_env(cls) -> "EasyAppOptions":
        """
        Build options from environment variables:

          ENABLE_LOGGING=true|false
          ENABLE_OBS=true|false
          LOG_LEVEL=DEBUG|INFO|...
          LOG_FORMAT=json|plain
          METRICS_PATH=/metrics
          OBS_SKIP_PATHS=/metrics,/health,/internal     (csv or space-separated)
        """
        return cls(
            logging=LoggingOptions(
                enable=_env_bool("ENABLE_LOGGING", True),
                level=_env_str("LOG_LEVEL"),
                fmt=_env_str("LOG_FORMAT"),
            ),
            observability=ObservabilityOptions(
                enable=_env_bool("ENABLE_OBS", True),
                metrics_path=_env_str("METRICS_PATH"),
                skip_metric_paths=_env_csv_paths("OBS_SKIP_PATHS"),
            ),
        )

    def merged_with(self, override: "EasyAppOptions | None") -> "EasyAppOptions":
        """
        Merge two option sets. Non-None fields in `override` win.
        (For iterables, if override provides a non-None value, it wins entirely.)
        """
        if override is None:
            return self

        # logging
        log = LoggingOptions(
            enable=(
                override.logging.enable
                if override.logging.enable is not None
                else self.logging.enable
            ),
            level=(
                override.logging.level if override.logging.level is not None else self.logging.level
            ),
            fmt=override.logging.fmt if override.logging.fmt is not None else self.logging.fmt,
        )

        # observability
        obs = ObservabilityOptions(
            enable=(
                override.observability.enable
                if override.observability.enable is not None
                else self.observability.enable
            ),
            db_engines=(
                override.observability.db_engines
                if override.observability.db_engines is not None
                else self.observability.db_engines
            ),
            metrics_path=(
                override.observability.metrics_path
                if override.observability.metrics_path is not None
                else self.observability.metrics_path
            ),
            skip_metric_paths=(
                override.observability.skip_metric_paths
                if override.observability.skip_metric_paths is not None
                else self.observability.skip_metric_paths
            ),
        )

        return EasyAppOptions(logging=log, observability=obs)


# ---------- Builders ----------


def easy_service_api(
    *,
    name: str,
    release: str,
    versions: Sequence[tuple[str | int, str, str | None]] | None = None,
    root_routers: list[str] | str | None = None,
    public_cors_origins: list[str] | str | None = None,
) -> FastAPI:
    """
    Pure app factory (no logging/obs side-effects).
    """
    service = ServiceInfo(name=name, release=release)
    specs = [
        APIVersionSpec(tag=str(tag), routers_package=pkg, public_base_url=base)
        for (tag, pkg, base) in (versions or [])
    ]
    return setup_service_api(
        service=service,
        versions=specs,
        root_title=name,
        root_routers=root_routers,
        public_cors_origins=public_cors_origins,
    )


def easy_service_app(
    *,
    name: str,
    release: str,
    versions: Sequence[tuple[str | int, str, str | None]] | None = None,
    root_routers: list[str] | str | None = None,
    public_cors_origins: list[str] | str | None = None,
    options: EasyAppOptions | None = None,
    # Back-compat shorthand flags (strongest precedence):
    enable_logging: bool | None = None,
    enable_observability: bool | None = None,
) -> FastAPI:
    """
    One-call bootstrap with env + options + flags:

      Precedence (strongest → weakest):
        1) enable_logging / enable_observability args
        2) `options=` object (per-field)
        3) `EasyAppOptions.from_env()`

      Env recognized:
        ENABLE_LOGGING=true|false
        ENABLE_OBS=true|false
        LOG_LEVEL=DEBUG|INFO|...
        LOG_FORMAT=json|plain
        METRICS_PATH=/metrics
        OBS_SKIP_PATHS=/metrics,/health,/internal
    """
    # 0) Start from env
    env_opts = EasyAppOptions.from_env()
    # 1) Merge explicit options on top
    effective = env_opts.merged_with(options)

    # 2) Apply strongest-prec flags
    if enable_logging is not None:
        effective.logging.enable = bool(enable_logging)
    if enable_observability is not None:
        effective.observability.enable = bool(enable_observability)

    # 3) Logging
    if effective.logging.enable:
        setup_logging(
            level=effective.logging.level
            or pick(
                prod=LogLevelOptions.INFO,
                test=LogLevelOptions.INFO,
                dev=LogLevelOptions.DEBUG,
                local=LogLevelOptions.DEBUG,
            ),
            fmt=effective.logging.fmt,  # None → auto (json in prod, plain elsewhere)
        )

    # 4) App
    app = easy_service_api(
        name=name,
        release=release,
        versions=versions,
        root_routers=root_routers,
        public_cors_origins=public_cors_origins,
    )

    # 5) Observability
    if effective.observability.enable:
        add_observability(
            app,
            db_engines=effective.observability.db_engines,
            metrics_path=effective.observability.metrics_path,
            skip_metric_paths=effective.observability.skip_metric_paths,
        )

    return app



from __future__ import annotations

from typing import Any, Sequence

from .repository import NoSqlRepository


class NoSqlService:
    """
    Small orchestration layer mirroring SqlService.
    """

    def __init__(self, repo: NoSqlRepository):
        self.repo = repo

    # hooks — override in subclasses if needed
    async def pre_create(self, data: dict[str, Any]) -> dict[str, Any]:
        return data

    async def pre_update(self, data: dict[str, Any]) -> dict[str, Any]:
        return data

    async def list(self, db, *, limit: int, offset: int, sort=None):
        return await self.repo.list(db, limit=limit, offset=offset, sort=sort)

    async def count(self, db) -> int:
        return await self.repo.count(db)

    async def get(self, db, id_value: Any):
        return await self.repo.get(db, id_value)

    async def create(self, db, data: dict[str, Any]):
        data = await self.pre_create(data)
        return await self.repo.create(db, data)

    async def update(self, db, id_value: Any, data: dict[str, Any]):
        data = await self.pre_update(data)
        return await self.repo.update(db, id_value, data)

    async def delete(self, db, id_value: Any) -> bool:
        return await self.repo.delete(db, id_value)

    async def search(
        self, db, *, q: str, fields: Sequence[str], limit: int, offset: int, sort=None
    ):
        return await self.repo.search(db, q=q, fields=fields, limit=limit, offset=offset, sort=sort)

    async def count_filtered(self, db, *, q: str, fields: Sequence[str]) -> int:
        return await self.repo.count_filtered(db, q=q, fields=fields)

    async def exists(self, db, *, where):
        return await self.repo.exists(db, where=where)



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Literal, Optional

from svc_infra.db.utils import normalize_dir, pascal, plural_snake, snake
from svc_infra.utils import ensure_init_py, render_template, write

_INIT_CONTENT_PAIRED = 'from . import documents, schemas\n\n__all__ = ["documents", "schemas"]\n'
_INIT_CONTENT_MINIMAL = "# package marker; add explicit exports here if desired\n"


def _ensure_init_py(dir_path: Path, overwrite: bool, paired: bool) -> Dict[str, Any]:
    content = _INIT_CONTENT_PAIRED if paired else _INIT_CONTENT_MINIMAL
    return ensure_init_py(dir_path, overwrite, paired, content)


# -------------- public API -------------------

Kind = Literal["entity"]


def scaffold_core(
    *,
    documents_dir: Path | str,
    schemas_dir: Path | str,
    entity_name: str = "Item",
    overwrite: bool = False,
    same_dir: bool = False,
    documents_filename: Optional[str] = None,
    schemas_filename: Optional[str] = None,
) -> Dict[str, Any]:
    """Create starter Mongo document model + CRUD schemas."""

    documents_dir = normalize_dir(documents_dir)
    schemas_dir = normalize_dir(documents_dir if same_dir else schemas_dir)

    ent = pascal(entity_name)
    coll = plural_snake(ent)

    documents_txt = render_template(
        tmpl_dir="svc_infra.db.nosql.mongo.templates",
        name="documents.py.tmpl",
        subs={"Entity": ent, "collection_name": coll},
    )
    schemas_txt = render_template(
        tmpl_dir="svc_infra.db.nosql.mongo.templates",
        name="schemas.py.tmpl",
        subs={"Entity": ent},  # (only if your schemas.tmpl doesn't need collection_name)
    )

    if same_dir:
        doc_path = documents_dir / "documents.py"
        sch_path = schemas_dir / "schemas.py"
    else:
        base = snake(entity_name)
        doc_path = documents_dir / (documents_filename or f"{base}.py")
        sch_path = schemas_dir / (schemas_filename or f"{base}.py")

    res_doc = write(doc_path, documents_txt, overwrite)
    res_sch = write(sch_path, schemas_txt, overwrite)

    init_results = []
    init_results.append(_ensure_init_py(documents_dir, overwrite, paired=same_dir))
    if schemas_dir != documents_dir:
        init_results.append(_ensure_init_py(schemas_dir, overwrite, paired=False))

    return {
        "status": "ok",
        "results": {"documents": res_doc, "schemas": res_sch, "inits": init_results},
    }


def scaffold_documents_core(
    *,
    dest_dir: Path | str,
    entity_name: str = "Item",
    overwrite: bool = False,
    documents_filename: Optional[str] = None,
) -> Dict[str, Any]:
    dest = normalize_dir(dest_dir)
    ent = pascal(entity_name)
    coll = plural_snake(ent)

    txt = render_template(
        tmpl_dir="svc_infra.db.nosql.mongo.templates",
        name="documents.py.tmpl",
        subs={"Entity": ent, "collection_name": coll},
    )
    filename = documents_filename or f"{snake(entity_name)}.py"
    res = write(dest / filename, txt, overwrite)
    _ensure_init_py(dest, overwrite, paired=False)
    return {"status": "ok", "result": res}


def scaffold_schemas_core(
    *,
    dest_dir: Path | str,
    entity_name: str = "Item",
    overwrite: bool = False,
    schemas_filename: Optional[str] = None,
) -> Dict[str, Any]:
    dest = normalize_dir(dest_dir)
    ent = pascal(entity_name)
    txt = render_template(
        tmpl_dir="svc_infra.db.nosql.mongo.templates", name="schemas.py.tmpl", subs={"Entity": ent}
    )
    filename = schemas_filename or f"{snake(entity_name)}.py"
    res = write(dest / filename, txt, overwrite)
    _ensure_init_py(dest, overwrite, paired=False)
    return {"status": "ok", "result": res}


def scaffold_resources_core(
    *,
    dest_dir: Path | str,
    entity_name: str = "Item",
    filename: Optional[str] = None,  # defaults to "resources.py"
    overwrite: bool = False,
) -> Dict[str, Any]:
    """
    Create a starter resources.py with:
      - empty RESOURCES list
      - empty index_builders() mapping
    Uses templates/scaffold/resources.py.tmpl (no hardcoded content).
    """
    # normalize
    dest = Path(dest_dir)
    if not dest.is_absolute():
        dest = (Path.cwd() / dest).resolve()
    dest.mkdir(parents=True, exist_ok=True)

    ent = pascal(entity_name)
    coll = plural_snake(ent)

    # render from file template
    txt = render_template(
        tmpl_dir="svc_infra.db.nosql.mongo.templates",
        name="resources.py.tmpl",
        subs={
            "Entity": ent,
            "collection_name": coll,
        },
    )

    # write file
    out_path = dest / (filename or "resources.py")
    res = write(out_path, txt, overwrite)

    # ensure __init__.py
    _ensure_init_py(dest, overwrite, paired=False)

    return {"status": "ok", "result": res}



from __future__ import annotations

import os
from typing import List, Optional

import typer

from svc_infra.db.sql.core import current as core_current
from svc_infra.db.sql.core import downgrade as core_downgrade
from svc_infra.db.sql.core import history as core_history
from svc_infra.db.sql.core import init_alembic as core_init_alembic
from svc_infra.db.sql.core import merge_heads as core_merge_heads
from svc_infra.db.sql.core import revision as core_revision
from svc_infra.db.sql.core import setup_and_migrate as core_setup_and_migrate
from svc_infra.db.sql.core import stamp as core_stamp
from svc_infra.db.sql.core import upgrade as core_upgrade


def apply_database_url(database_url: Optional[str]) -> None:
    """If provided, set SQL_URL for the current process."""
    if database_url:
        os.environ["SQL_URL"] = database_url


def cmd_init(
    database_url: Optional[str] = typer.Option(
        None,
        help="Database URL; overrides env SQL_URL for this command. "
        "Async vs sync is auto-detected from the URL.",
    ),
    discover_packages: Optional[List[str]] = typer.Option(
        None,
        help="Packages to search for SQLAlchemy metadata; may pass multiple. "
        "If omitted, automatic discovery is used.",
    ),
    overwrite: bool = typer.Option(False, help="Overwrite existing files if present."),
):
    """
    Initialize Alembic scaffold. The env.py variant (async vs. sync) is
    auto-detected from SQL_URL (if available at init time).
    """
    apply_database_url(database_url)
    core_init_alembic(
        discover_packages=discover_packages,
        overwrite=overwrite,
    )


def cmd_revision(
    message: str = typer.Option(..., "-m", "--message", help="Revision message."),
    database_url: Optional[str] = typer.Option(
        None, help="Database URL; overrides env for this command."
    ),
    autogenerate: bool = typer.Option(False, help="Autogenerate migrations by comparing metadata."),
    head: Optional[str] = typer.Option("head", help="Set the head to base this revision on."),
    branch_label: Optional[str] = typer.Option(None, help="Branch label."),
    version_path: Optional[str] = typer.Option(None, help="Alternative versions/ path."),
    sql: bool = typer.Option(False, help="Don't generate Python; dump SQL to stdout."),
):
    """Create a new Alembic revision, either empty or autogenerated."""
    apply_database_url(database_url)
    core_revision(
        message=message,
        autogenerate=autogenerate,
        head=head,
        branch_label=branch_label,
        version_path=version_path,
        sql=sql,
    )


def cmd_upgrade(
    revision_target: str = typer.Argument("head", help="Target revision (default head)."),
    database_url: Optional[str] = typer.Option(
        None, help="Database URL; overrides env for this command."
    ),
):
    """Upgrade to a later version."""
    apply_database_url(database_url)
    core_upgrade(revision_target=revision_target)


def cmd_downgrade(
    revision_target: str = typer.Argument("-1", help="Target revision (default -1)."),
    database_url: Optional[str] = typer.Option(
        None, help="Database URL; overrides env for this command."
    ),
):
    """Downgrade to an earlier version."""
    apply_database_url(database_url)
    core_downgrade(revision_target=revision_target)


def cmd_current(
    database_url: Optional[str] = typer.Option(
        None, help="Database URL; overrides env for this command."
    ),
    verbose: bool = typer.Option(False, help="Verbose output."),
):
    """Display the current revision for each database."""
    apply_database_url(database_url)
    core_current(verbose=verbose)


def cmd_history(
    database_url: Optional[str] = typer.Option(
        None, help="Database URL; overrides env for this command."
    ),
    verbose: bool = typer.Option(False, help="Verbose output."),
):
    """List changeset scripts in chronological order."""
    apply_database_url(database_url)
    core_history(verbose=verbose)


def cmd_stamp(
    revision_target: str = typer.Argument("head"),
    database_url: Optional[str] = typer.Option(
        None, help="Database URL; overrides env for this command."
    ),
):
    """Stamp the revision table with the given revision; don't run migrations."""
    apply_database_url(database_url)
    core_stamp(revision_target=revision_target)


def cmd_merge_heads(
    database_url: Optional[str] = typer.Option(
        None, help="Database URL; overrides env for this command."
    ),
    message: Optional[str] = typer.Option(None, "-m", "--message", help="Merge revision message."),
):
    """Create a merge revision for multiple heads."""
    apply_database_url(database_url)
    core_merge_heads(message=message)


def cmd_setup_and_migrate(
    database_url: Optional[str] = typer.Option(
        None,
        help="Overrides env for this command. Async vs sync is auto-detected from the URL.",
    ),
    overwrite_scaffold: bool = typer.Option(False, help="Overwrite alembic scaffold if present."),
    create_db_if_missing: bool = typer.Option(True, help="Create the database/schema if missing."),
    create_followup_revision: bool = typer.Option(
        True, help="Create an autogen follow-up revision if revisions already exist."
    ),
    initial_message: str = typer.Option("initial schema"),
    followup_message: str = typer.Option("autogen"),
):
    """
    End-to-end: ensure DB exists, scaffold Alembic, create/upgrade revisions.
    Async vs. sync is inferred from SQL_URL.
    """
    apply_database_url(database_url)
    core_setup_and_migrate(
        overwrite_scaffold=overwrite_scaffold,
        create_db_if_missing=create_db_if_missing,
        create_followup_revision=create_followup_revision,
        initial_message=initial_message,
        followup_message=followup_message,
    )


def register(app: typer.Typer) -> None:
    app.command("sql-init")(cmd_init)
    app.command("sql-revision")(cmd_revision)
    app.command("sql-upgrade")(cmd_upgrade)
    app.command("sql-downgrade")(cmd_downgrade)
    app.command("sql-current")(cmd_current)
    app.command("sql-history")(cmd_history)
    app.command("sql-stamp")(cmd_stamp)
    app.command("sql-merge-heads")(cmd_merge_heads)
    app.command("sql-setup-and-migrate")(cmd_setup_and_migrate)



"""
Cache tag invalidation utilities.

This module provides functionality for invalidating cache entries by tags
with fallback strategies for different cashews versions.
"""

import logging

from cashews import cache as _cache

logger = logging.getLogger(__name__)


async def invalidate_tags(*tags: str) -> int:
    """
    Invalidate cache entries by tags with fallback strategies.

    This function tries multiple approaches to invalidate cache tags,
    providing compatibility across different cashews versions.

    Args:
        *tags: Cache tags to invalidate

    Returns:
        Number of invalidated entries (best effort)
    """
    if not tags:
        return 0

    count = 0

    # Strategy 1: Modern cashews invalidate with tags parameter
    try:
        result = await _cache.invalidate(tags=list(tags))
        return int(result) if isinstance(result, int) else len(tags)
    except (TypeError, AttributeError):
        pass
    except Exception as e:
        logger.warning(f"Modern tag invalidation failed: {e}")

    # Strategy 2: Legacy cashews invalidate with positional args
    try:
        result = await _cache.invalidate(*tags)
        return int(result) if isinstance(result, int) else len(tags)
    except (TypeError, AttributeError):
        pass
    except Exception as e:
        logger.warning(f"Legacy tag invalidation failed: {e}")

    # Strategy 3: Individual tag methods
    for tag in tags:
        for method_name in ("delete_tag", "invalidate_tag", "tag_invalidate"):
            if hasattr(_cache, method_name):
                try:
                    method = getattr(_cache, method_name)
                    result = await method(tag)
                    count += int(result) if isinstance(result, int) else 1
                    break
                except Exception as e:
                    logger.debug(f"Tag method {method_name} failed for tag {tag}: {e}")
                    continue
        else:
            # Strategy 4: Pattern matching fallback
            for method_name in ("delete_match", "invalidate_match", "invalidate"):
                if hasattr(_cache, method_name):
                    try:
                        method = getattr(_cache, method_name)
                        pattern = f"*{tag}*"
                        result = await method(pattern)
                        count += int(result) if isinstance(result, int) else 1
                        break
                    except Exception as e:
                        logger.debug(f"Pattern method {method_name} failed for tag {tag}: {e}")
                        continue

    return count



from __future__ import annotations

from typing import Any, Mapping, Optional

from sqlalchemy.engine import Engine

try:
    from sqlalchemy.ext.asyncio import AsyncEngine
except Exception:  # optional
    AsyncEngine = None  # type: ignore

from .base import counter, gauge

_pool_in_use = gauge(
    "db_pool_in_use",
    "Checked-out connections",
    labels=["db"],
    multiprocess_mode="livesum",
)
_pool_available = gauge(
    "db_pool_available",
    "Available idle connections",
    labels=["db"],
    multiprocess_mode="livesum",
)
_pool_checked_out_total = counter("db_pool_checkedout_total", "Total checkouts", labels=["db"])
_pool_checked_in_total = counter("db_pool_checkedin_total", "Total checkins", labels=["db"])


def _label(labels: Optional[Mapping[str, str]]) -> str:
    return (labels or {}).get("db", "default")


def bind_sqlalchemy_pool_metrics(
    engine: Engine | Any, labels: Optional[Mapping[str, str]] = None
) -> None:
    """Bind event listeners for pool metrics. Works for sync Engine.
    For AsyncEngine pass engine.sync_engine."""
    label = _label(labels)
    sync_engine: Engine = getattr(engine, "sync_engine", engine)

    from sqlalchemy import event

    @event.listens_for(sync_engine, "engine_connect")
    def _(conn, branch):  # noqa
        # Update gauges on engine_connect as a cheap heartbeat
        pool = sync_engine.pool
        try:
            _pool_in_use.labels(label).set(pool.checkedout())
            _pool_available.labels(label).set(pool.size() - pool.checkedout())
        except Exception:
            pass

    @event.listens_for(sync_engine, "checkout")
    def _checkout(dbapi_con, con_record, con_proxy):  # noqa
        _pool_checked_out_total.labels(label).inc()
        try:
            pool = sync_engine.pool
            _pool_in_use.labels(label).set(pool.checkedout())
            _pool_available.labels(label).set(pool.size() - pool.checkedout())
        except Exception:
            pass

    @event.listens_for(sync_engine, "checkin")
    def _checkin(dbapi_con, con_record):  # noqa
        _pool_checked_in_total.labels(label).inc()
        try:
            pool = sync_engine.pool
            _pool_in_use.labels(label).set(pool.checkedout())
            _pool_available.labels(label).set(pool.size() - pool.checkedout())
        except Exception:
            pass



from __future__ import annotations

from typing import Any, Sequence

from fastapi import HTTPException
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from svc_infra.db.sql.repository import SqlRepository


class SqlService:
    """
    Small orchestration layer. Extend/override for business logic, RBAC, etc.
    """

    def __init__(self, repo: SqlRepository):
        self.repo = repo

    # hooks — override in subclasses if needed
    async def pre_create(self, data: dict[str, Any]) -> dict[str, Any]:
        return data

    async def pre_update(self, data: dict[str, Any]) -> dict[str, Any]:
        return data

    async def list(self, session: AsyncSession, *, limit: int, offset: int, order_by=None):
        return await self.repo.list(session, limit=limit, offset=offset, order_by=order_by)

    async def count(self, session: AsyncSession) -> int:
        return await self.repo.count(session)

    async def get(self, session: AsyncSession, id_value: Any):
        return await self.repo.get(session, id_value)

    async def create(self, session: AsyncSession, data: dict[str, Any]):
        data = await self.pre_create(data)
        try:
            return await self.repo.create(session, data)
        except IntegrityError as e:
            # unique constraint or not-null -> 409/400 instead of 500
            msg = str(e.orig) if getattr(e, "orig", None) else str(e)
            if "duplicate key value" in msg or "UniqueViolation" in msg:
                raise HTTPException(status_code=409, detail="Record already exists.") from e
            if "not-null" in msg or "NotNullViolation" in msg:
                raise HTTPException(status_code=400, detail="Missing required field.") from e
            raise  # unknown, let your error middleware turn into 500

    async def update(self, session: AsyncSession, id_value: Any, data: dict[str, Any]):
        data = await self.pre_update(data)
        return await self.repo.update(session, id_value, data)

    async def delete(self, session: AsyncSession, id_value: Any) -> bool:
        return await self.repo.delete(session, id_value)

    async def search(
        self,
        session: AsyncSession,
        *,
        q: str,
        fields: Sequence[str],
        limit: int,
        offset: int,
        order_by=None,
    ):
        return await self.repo.search(
            session, q=q, fields=fields, limit=limit, offset=offset, order_by=order_by
        )

    async def count_filtered(self, session: AsyncSession, *, q: str, fields: Sequence[str]) -> int:
        return await self.repo.count_filtered(session, q=q, fields=fields)

    async def exists(self, session: AsyncSession, *, where):
        return await self.repo.exists(session, where=where)



from __future__ import annotations

import asyncio
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Sequence, Set, Union

from alembic.config import Config
from dotenv import load_dotenv
from sqlalchemy import inspect, text
from sqlalchemy.engine import URL, make_url
from sqlalchemy.exc import DBAPIError, OperationalError

from .constants import ASYNC_DRIVER_HINT, DEFAULT_DB_ENV_VARS

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine as SyncEngine
    from sqlalchemy.ext.asyncio import AsyncEngine as AsyncEngineType
else:
    SyncEngine = Any  # type: ignore
    AsyncEngineType = Any  # type: ignore

try:
    # Runtime import (may be missing if async extras aren’t installed)
    from sqlalchemy.ext.asyncio import create_async_engine as _create_async_engine  # type: ignore
except Exception:  # pragma: no cover - optional dep
    _create_async_engine = None  # type: ignore

try:
    from sqlalchemy import create_engine as _create_engine  # type: ignore
except Exception:  # pragma: no cover - optional env
    _create_engine = None  # type: ignore


def prepare_process_env(
    project_root: Path | str,
    discover_packages: Optional[Sequence[str]] = None,
) -> None:
    """
    Prepare process environment so Alembic can import the project cleanly.

    Notes:
        - Does NOT set SQL_URL (expect it to be set in your .env / environment).
        - Discovery is automatic via env.py. 'discover_packages' is kept for
          backward-compat only; prefer leaving it None.
    """
    root = Path(project_root).resolve()
    load_dotenv(dotenv_path=root / ".env", override=False)
    os.environ.setdefault("SKIP_APP_INIT", "1")

    # Make <project>/src importable (env.py also handles this defensively)
    src_dir = root / "src"
    if src_dir.exists():
        sys_path = os.environ.get("PYTHONPATH", "")
        parts = [str(src_dir)] + ([sys_path] if sys_path else [])
        os.environ["PYTHONPATH"] = os.pathsep.join(parts)

    # Optional override (discouraged—automatic discovery is preferred)
    if discover_packages:
        os.environ["ALEMBIC_DISCOVER_PACKAGES"] = ",".join(discover_packages)


def _read_secret_from_file(path: str) -> Optional[str]:
    """Return file contents if path exists, else None."""
    try:
        p = Path(path)
        if p.exists():
            return p.read_text(encoding="utf-8").strip()
    except Exception:
        pass
    return None


def _compose_url_from_parts() -> Optional[str]:
    """
    Compose a SQLAlchemy URL from component env vars.
    Supports private DNS hostnames and Unix sockets.

    Recognized envs:
      DB_DIALECT (default: postgresql), DB_DRIVER (optional, e.g. asyncpg, psycopg),
      DB_HOST (hostname or Unix socket dir), DB_PORT,
      DB_NAME, DB_USER, DB_PASSWORD,
      DB_PARAMS (raw query string like 'sslmode=require&connect_timeout=5')
    """
    dialect = os.getenv("DB_DIALECT", "").strip() or "postgresql"
    driver = os.getenv("DB_DRIVER", "").strip()  # e.g. asyncpg, psycopg, pymysql, aiosqlite
    host = os.getenv("DB_HOST", "").strip() or None
    port = os.getenv("DB_PORT", "").strip() or None
    db = os.getenv("DB_NAME", "").strip() or None
    user = os.getenv("DB_USER", "").strip() or None
    pwd = os.getenv("DB_PASSWORD", "").strip() or None
    params = os.getenv("DB_PARAMS", "").strip() or ""

    if not (host and db):
        return None

    # Build SQLAlchemy URL safely
    drivername = f"{dialect}+{driver}" if driver else dialect
    query = dict(q.split("=", 1) for q in params.split("&") if q) if params else {}

    # URL.create handles unix socket paths when host begins with a slash
    try:
        url = URL.create(
            drivername=drivername,
            username=user or None,
            password=pwd or None,
            host=host if (host and not host.startswith("/")) else None,
            port=int(port) if (port and port.isdigit()) else None,
            database=db,
            query=query,
        )
        # If host is a unix socket dir, place it in query as host param many drivers understand
        if host and host.startswith("/"):
            # e.g. for psycopg/psycopg2: host=/cloudsql/instance; for MySQL: unix_socket=/path
            if "postgresql" in drivername:
                url = url.set(query={**url.query, "host": host})
            elif "mysql" in drivername:
                url = url.set(query={**url.query, "unix_socket": host})
        return str(url)
    except Exception:
        return None


# ---------- Environment helpers ----------


def get_database_url_from_env(
    required: bool = True,
    env_vars: Sequence[str] = DEFAULT_DB_ENV_VARS,
) -> Optional[str]:
    """
    Resolve the database connection string, with support for:
      - Primary env vars (in order): DEFAULT_DB_ENV_VARS (e.g. SQL_URL, PRIVATE_SQL_URL, DB_URL).
      - Secret file pointers: <NAME>_FILE (reads file contents).
      - Well-known locations: SQL_URL_FILE, /run/secrets/database_url.
      - Composed from parts: DB_* (host, port, name, user, password, params).
    When a value is found, it is also written back into os.environ["SQL_URL"] for downstream code.
    """
    # Load .env without clobbering existing process env
    load_dotenv(override=False)

    # 1) Direct envs (and support "file:" or absolute path values)
    for key in env_vars:
        val = os.getenv(key)
        if val and val.strip():
            s = val.strip()
            # Some platforms inject a file-pointer-like value
            if s.startswith("file:"):
                s = s[5:]
            if os.path.isabs(s) and Path(s).exists():
                file_val = _read_secret_from_file(s)
                if file_val:
                    os.environ["SQL_URL"] = file_val
                    return file_val
            os.environ["SQL_URL"] = s
            return s

        # Companion NAME_FILE secret path (e.g., SQL_URL_FILE)
        file_key = f"{key}_FILE"
        file_path = os.getenv(file_key)
        if file_path:
            file_val = _read_secret_from_file(file_path)
            if file_val:
                os.environ["SQL_URL"] = file_val
                return file_val

    # 2) Conventional secret envs
    file_path = os.getenv("SQL_URL_FILE")
    if file_path:
        file_val = _read_secret_from_file(file_path)
        if file_val:
            os.environ["SQL_URL"] = file_val
            return file_val

    # 3) Docker/K8s default secret mount
    file_val = _read_secret_from_file("/run/secrets/database_url")
    if file_val:
        os.environ["SQL_URL"] = file_val
        return file_val

    # 4) Compose from parts (DB_DIALECT/DB_DRIVER/DB_HOST/.../DB_PARAMS)
    composed = _compose_url_from_parts()
    if composed:
        os.environ["SQL_URL"] = composed
        return composed

    if required:
        raise RuntimeError(
            "Database URL not set. Set SQL_URL (or PRIVATE_SQL_URL / DB_URL), "
            "or provide DB_* parts (DB_HOST, DB_NAME, etc.), or a *_FILE secret."
        )
    return None


def _ensure_timeout_default(u: URL) -> URL:
    """
    Ensure a conservative connection timeout is present for libpq-based drivers.
    For psycopg/psycopg2, 'connect_timeout' is honored via the query string.
    """
    backend = (u.get_backend_name() or "").lower()
    if backend not in ("postgresql", "postgres"):
        return u
    if "connect_timeout" in u.query:
        return u
    # Default 10s unless overridden
    t = int(os.getenv("DB_CONNECT_TIMEOUT", "10"))
    return u.set(query={**u.query, "connect_timeout": str(t)})


# ---------- URL utilities ----------


def is_async_url(url: URL | str) -> bool:
    u = make_url(url) if isinstance(url, str) else url
    dn = u.drivername or ""
    return bool(ASYNC_DRIVER_HINT.search(dn))


def with_database(url: URL | str, database: Optional[str]) -> URL:
    """Return a copy of URL with the database name replaced.

    Works for most dialects. For SQLite/DuckDB file URLs, `database` is the file path.
    """
    u = make_url(url) if isinstance(url, str) else url
    return u.set(database=database)


# ---------- Engine creation ----------


def _coerce_to_async_url(url: str) -> str:
    """Coerce common sync driver URLs to async-capable URLs.

    - postgresql:// or postgres://        -> postgresql+asyncpg://
    - postgresql+psycopg2:// or +psycopg  -> postgresql+asyncpg://
    - mysql:// or mysql+pymysql://        -> mysql+aiomysql://
    - sqlite://                           -> sqlite+aiosqlite://
    If already async (contains +asyncpg/+aiomysql/+aiosqlite), leave unchanged.
    """
    low = url.lower()
    if "+asyncpg" in low or "+aiomysql" in low or "+aiosqlite" in low:
        return url
    if low.startswith("postgresql+psycopg2://"):
        return "postgresql+asyncpg://" + url.split("://", 1)[1]
    if low.startswith("postgresql+psycopg://"):
        return "postgresql+asyncpg://" + url.split("://", 1)[1]
    if low.startswith("postgresql://"):
        return "postgresql+asyncpg://" + url.split("://", 1)[1]
    if low.startswith("postgres://"):
        return "postgresql+asyncpg://" + url.split("://", 1)[1]
    if low.startswith("mysql+pymysql://") or low.startswith("mysql://"):
        return "mysql+aiomysql://" + url.split("://", 1)[1]
    if low.startswith("sqlite://") and not low.startswith("sqlite+aiosqlite://"):
        return "sqlite+aiosqlite://" + url.split("://", 1)[1]
    return url


def _is_mod_available(name: str) -> bool:
    try:
        __import__(name)
        return True
    except Exception:
        return False


def _coerce_sync_driver(u: URL) -> URL:
    """
    If URL has no explicit sync driver, pick one that’s available.
    Postgres preference: psycopg (v3) → psycopg2
    Optional override: DB_FORCE_DRIVER=psycopg|psycopg2
    """
    dn = (u.drivername or "").lower()
    if "+" in dn:
        return u

    backend = (u.get_backend_name() or "").lower()
    if backend in ("postgresql", "postgres"):
        force = (os.getenv("DB_FORCE_DRIVER") or "").strip().lower()
        if force in {"psycopg", "psycopg2"}:
            return u.set(drivername=f"postgresql+{force}")

        # prefer psycopg v3 if present; else psycopg2; else leave bare (SA default)
        if _is_mod_available("psycopg"):
            return u.set(drivername="postgresql+psycopg")
        if _is_mod_available("psycopg2"):
            return u.set(drivername="postgresql+psycopg2")
        return u

    if backend == "mysql":
        if _is_mod_available("pymysql"):
            return u.set(drivername="mysql+pymysql")
        return u

    if backend == "sqlite":
        return u

    if backend in ("mssql",):
        if _is_mod_available("pyodbc"):
            return u.set(drivername="mssql+pyodbc")
        return u

    return u


def _coerce_pg_maintenance_driver(u: URL) -> URL:
    """
    Ensure the maintenance connection for Postgres uses a sync driver that is installed,
    because ensure_database_exists() often needs a *sync* connection for CREATE DATABASE.
    """
    # If async → leave to async branch; this is for sync path only.
    if "+" in (u.drivername or ""):
        # If explicit async driver was given, leave it (caller decides).
        return u
    backend = (u.get_backend_name() or "").lower()
    if backend in ("postgresql", "postgres"):
        # prefer psycopg, then psycopg2
        if _is_mod_available("psycopg"):
            return u.set(drivername="postgresql+psycopg")
        if _is_mod_available("psycopg2"):
            return u.set(drivername="postgresql+psycopg2")
    return u


def _ensure_ssl_default(u: URL) -> URL:
    backend = (u.get_backend_name() or "").lower()
    if backend not in ("postgresql", "postgres"):
        return u

    driver = (u.drivername or "").lower()

    # If any SSL hint already present, do nothing
    if any(k in u.query for k in ("sslmode", "ssl", "sslrootcert", "sslcert", "sslkey")):
        return u

    # Allow env override; support both common spellings
    mode_env = os.getenv("DB_SSLMODE_DEFAULT") or os.getenv("PGSSLMODE") or os.getenv("PGSSL_MODE")
    mode = (mode_env or "").strip()

    if "+asyncpg" in driver:
        # asyncpg: use 'ssl=true' (SQLAlchemy forwards to asyncpg)
        if "ssl" not in u.query:
            return u.set(query={**u.query, "ssl": "true"})
        return u
    else:
        # libpq-based drivers: use sslmode (default 'require' for hosted PG)
        mode = mode or "require"
        return u.set(query={**u.query, "sslmode": mode})


def _ensure_ssl_default_async(u: URL) -> URL:
    backend = (u.get_backend_name() or "").lower()
    if backend in ("postgresql", "postgres"):
        # asyncpg prefers 'ssl=true' via SQLAlchemy param; if already present, keep it
        if any(k in u.query for k in ("ssl", "sslmode", "sslrootcert", "sslcert", "sslkey")):
            return u
        return u.set(query={**u.query, "ssl": "true"})
    return u


def _certifi_ca() -> str | None:
    try:
        import certifi

        return certifi.where()
    except Exception:
        return None


def build_engine(url: URL | str, echo: bool = False) -> Union[SyncEngine, AsyncEngineType]:
    u = make_url(url) if isinstance(url, str) else url

    # Keep your existing PG helpers
    u = _ensure_ssl_default(u)
    u = _ensure_timeout_default(u)

    connect_args: dict[str, Any] = {}

    # ----------------- ASYNC -----------------
    if is_async_url(u):
        if _create_async_engine is None:
            raise RuntimeError(
                "Async driver URL provided but SQLAlchemy async extras are not available."
            )

        # asyncpg: honor connection timeout
        if "+asyncpg" in (u.drivername or ""):
            connect_args["timeout"] = int(os.getenv("DB_CONNECT_TIMEOUT", "10"))

        # NEW: aiomysql SSL default
        if "+aiomysql" in (u.drivername or "") and not any(
            k in u.query for k in ("ssl", "ssl_ca", "sslmode")
        ):
            # aiomysql accepts an ssl.SSLContext or True
            try:
                import ssl

                ca = _certifi_ca()
                ctx = ssl.create_default_context(cafile=ca) if ca else ssl.create_default_context()
                # if your host uses a public CA, verification works;
                # if not, you can relax verification (not recommended):
                #   ctx.check_hostname = False
                #   ctx.verify_mode = ssl.CERT_NONE
                connect_args["ssl"] = ctx
            except Exception:
                connect_args["ssl"] = True  # minimal hint to enable TLS

        kwargs: dict[str, Any] = {"echo": echo, "pool_pre_ping": True}
        if connect_args:
            kwargs["connect_args"] = connect_args
        return _create_async_engine(u, **kwargs)

    # ----------------- SYNC -----------------
    u = _coerce_sync_driver(u)
    if _create_engine is None:
        raise RuntimeError("SQLAlchemy create_engine is not available in this environment.")

    dn = (u.drivername or "").lower()

    # psycopg v3 quirk (optional)
    if dn.startswith("postgresql+psycopg") and u.password:
        connect_args["password"] = u.password

    # NEW: pymysql SSL default
    if dn.startswith("mysql+pymysql") and not any(
        k in u.query for k in ("ssl", "ssl_ca", "sslmode")
    ):
        ca = _certifi_ca()
        if ca:
            # PyMySQL expects a dict of SSL args; 'ca' is the common one
            connect_args["ssl"] = {"ca": ca}
        else:
            # Fallback: empty dict enables TLS without CA pinning (works on many hosts)
            connect_args["ssl"] = {}

        # Optional: if your provider requires it, you can also add:
        # connect_args.setdefault("client_flag", 0)

    kwargs: dict[str, Any] = {"echo": echo, "pool_pre_ping": True}
    if connect_args:
        kwargs["connect_args"] = connect_args
    return _create_engine(u, **kwargs)


# ---------- Identifier quoting helpers ----------


def _pg_quote_ident(name: str) -> str:
    """
    Escape embedded double quotes for PostgreSQL identifiers.
    Caller must wrap with double quotes.
    """
    if name is None:
        raise ValueError("Identifier cannot be None")
    return name.replace('"', '""')


def _mysql_quote_ident(name: str) -> str:
    """
    Escape embedded backticks for MySQL/MariaDB identifiers.
    Caller must wrap with backticks.
    """
    if name is None:
        raise ValueError("Identifier cannot be None")
    return name.replace("`", "``")


# ---------- Database bootstrap (per-dialect) ----------


async def _pg_create_database_async(url: URL) -> None:
    assert is_async_url(url)
    target_db = url.database
    if not target_db:
        return

    u = _ensure_ssl_default(make_url(url))
    maintenance_url = with_database(u, "postgres")
    engine: AsyncEngineType = build_engine(maintenance_url)  # type: ignore[assignment]

    try:
        async with engine.begin() as conn:
            exists = await conn.scalar(
                text("SELECT 1 AS one FROM pg_database WHERE datname = :name"),
                {"name": target_db},
            )
            if not exists:
                quoted = _pg_quote_ident(target_db)
                await conn.execution_options(isolation_level="AUTOCOMMIT").execute(
                    text(f'CREATE DATABASE "{quoted}"')
                )
    except DBAPIError as e:
        if "permission denied" in str(e).lower():
            pass
        else:
            raise
    finally:
        await engine.dispose()


def _pg_create_database_sync(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return

    # Build a maintenance URL pointing at the 'postgres' DB
    u = _ensure_ssl_default(make_url(url))
    maintenance_url = with_database(u, "postgres")

    # Make sure it has an installed *sync* driver (psycopg or psycopg2)
    maintenance_url = _coerce_pg_maintenance_driver(maintenance_url)

    # Try connecting to 'postgres'; if that fails (some hosts restrict it), try 'template1'
    try:
        engine: SyncEngine = build_engine(maintenance_url)  # type: ignore[assignment]
    except Exception:
        alt_url = with_database(maintenance_url, "template1")
        engine = build_engine(alt_url)  # type: ignore[assignment]

    try:
        with engine.begin() as conn:
            exists = conn.scalar(
                text("SELECT 1 FROM pg_database WHERE datname = :name"),
                {"name": target_db},
            )
            if not exists:
                quoted = _pg_quote_ident(target_db)
                conn.execution_options(isolation_level="AUTOCOMMIT").execute(
                    text(f'CREATE DATABASE "{quoted}"')
                )
    except DBAPIError as e:
        # If permission error, log and continue (DB likely pre-provisioned on the host)
        if "permission denied" in str(e).lower():
            pass
        else:
            raise
    finally:
        engine.dispose()


async def _mysql_create_database_async(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    engine: AsyncEngineType = build_engine(base_url)  # type: ignore[assignment]
    async with engine.begin() as conn:
        exists = await conn.scalar(
            text("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = :name"),
            {"name": target_db},
        )
        if not exists:
            quoted = _mysql_quote_ident(target_db)
            await conn.execute(text(f"CREATE DATABASE `{quoted}`"))
    await engine.dispose()


def _mysql_create_database_sync(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    engine: SyncEngine = build_engine(base_url)  # type: ignore[assignment]
    with engine.begin() as conn:
        exists = conn.scalar(
            text("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = :name"),
            {"name": target_db},
        )
        if not exists:
            quoted = _mysql_quote_ident(target_db)
            conn.execute(text(f"CREATE DATABASE `{quoted}`"))
    engine.dispose()


def _sqlite_prepare_filesystem(url: URL) -> None:
    # file-based sqlite path e.g., sqlite:////tmp/file.db or sqlite+pysqlite:////path
    database = url.database
    if not database or database in {":memory:", "memory:"}:
        return
    try:
        path = Path(database)
    except Exception:
        return
    if path.parent and not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)


# ---- Extra dialect helpers (best-effort) ------------------------------------


def _duckdb_prepare_filesystem(url: URL) -> None:
    # duckdb:///path/to/file.duckdb (or :memory:)
    database = url.database
    if not database or database in {":memory:", "memory:"}:
        return
    try:
        path = Path(database)
    except Exception:
        return
    if path.parent and not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)


def _cockroach_create_database_sync(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    eng: SyncEngine = build_engine(base_url)  # type: ignore[assignment]
    try:
        with eng.begin() as conn:
            conn.execute(text(f'CREATE DATABASE IF NOT EXISTS "{target_db}"'))
    finally:
        eng.dispose()


async def _cockroach_create_database_async(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    engine: AsyncEngineType = build_engine(base_url)  # type: ignore[assignment]
    try:
        async with engine.begin() as conn:
            await conn.execute(text(f'CREATE DATABASE IF NOT EXISTS "{target_db}"'))
    finally:
        await engine.dispose()


def _mssql_create_database_sync(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    master_url = with_database(url, "master")
    eng: SyncEngine = build_engine(master_url)  # type: ignore[assignment]
    try:
        with eng.begin() as conn:
            exists = conn.scalar(
                text("SELECT 1 AS one FROM sys.databases WHERE name = :name"),
                {"name": target_db},
            )
            if not exists:
                conn.execute(text(f"CREATE DATABASE [{target_db}]"))
    finally:
        eng.dispose()


async def _mssql_create_database_async(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    master_url = with_database(url, "master")
    engine: AsyncEngineType = build_engine(master_url)  # type: ignore[assignment]
    try:
        async with engine.begin() as conn:
            exists = await conn.scalar(
                text("SELECT 1 AS one FROM sys.databases WHERE name = :name"),
                {"name": target_db},
            )
            if not exists:
                await conn.execute(text(f"CREATE DATABASE [{target_db}]"))
    finally:
        await engine.dispose()


def _snowflake_create_database_sync(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    eng: SyncEngine = build_engine(base_url)  # type: ignore[assignment]
    try:
        with eng.begin() as conn:
            conn.execute(text(f'CREATE DATABASE IF NOT EXISTS "{target_db}"'))
    finally:
        eng.dispose()


async def _snowflake_create_database_async(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    engine: AsyncEngineType = build_engine(base_url)  # type: ignore[assignment]
    try:
        async with engine.begin() as conn:
            await conn.execute(text(f'CREATE DATABASE IF NOT EXISTS "{target_db}"'))
    finally:
        await engine.dispose()


def _redshift_create_database_sync(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    try:
        eng: SyncEngine = build_engine(base_url)  # type: ignore[assignment]
    except Exception:
        eng = build_engine(with_database(url, "dev"))  # type: ignore[assignment]
    try:
        with eng.begin() as conn:
            exists = conn.scalar(
                text("SELECT 1 AS one FROM pg_database WHERE datname = :name"),
                {"name": target_db},
            )
            if not exists:
                conn.execute(text(f'CREATE DATABASE "{target_db}"'))
    finally:
        eng.dispose()


async def _redshift_create_database_async(url: URL) -> None:
    target_db = url.database
    if not target_db:
        return
    base_url = with_database(url, None)
    try:
        engine: AsyncEngineType = build_engine(base_url)  # type: ignore[assignment]
    except Exception:
        engine = build_engine(with_database(url, "dev"))  # type: ignore[assignment]
    try:
        async with engine.begin() as conn:
            exists = await conn.scalar(
                text("SELECT 1 AS one FROM pg_database WHERE datname = :name"),
                {"name": target_db},
            )
            if not exists:
                await conn.execute(text(f'CREATE DATABASE "{target_db}"'))
    finally:
        await engine.dispose()


# ---------- Entry: ensure database ----------


def ensure_database_exists(url: URL | str) -> None:
    if url is None:
        raise RuntimeError(
            "ensure_database_exists: database URL is required but None was provided."
        )
    u = make_url(url) if isinstance(url, str) else url
    backend = (u.get_backend_name() or "").lower()

    if backend.startswith("sqlite"):
        _sqlite_prepare_filesystem(u)
        return
    if backend.startswith("duckdb"):
        _duckdb_prepare_filesystem(u)
        return

    if backend.startswith(("postgresql", "postgres")):
        return (
            asyncio.run(_pg_create_database_async(u))
            if is_async_url(u)
            else _pg_create_database_sync(u)
        )
    if backend.startswith(("mysql", "mariadb")):
        return (
            asyncio.run(_mysql_create_database_async(u))
            if is_async_url(u)
            else _mysql_create_database_sync(u)
        )
    if backend.startswith(("cockroach", "cockroachdb")):
        return (
            asyncio.run(_cockroach_create_database_async(u))
            if is_async_url(u)
            else _cockroach_create_database_sync(u)
        )
    if backend.startswith("mssql"):
        return (
            asyncio.run(_mssql_create_database_async(u))
            if is_async_url(u)
            else _mssql_create_database_sync(u)
        )
    if backend.startswith("snowflake"):
        return (
            asyncio.run(_snowflake_create_database_async(u))
            if is_async_url(u)
            else _snowflake_create_database_sync(u)
        )
    if backend.startswith("redshift"):
        return (
            asyncio.run(_redshift_create_database_async(u))
            if is_async_url(u)
            else _redshift_create_database_sync(u)
        )

    # Fallback: just ping
    try:
        eng = build_engine(u)
        if is_async_url(u):

            async def _ping_and_dispose():
                async with eng.begin() as conn:  # type: ignore[call-arg]
                    await conn.execute(text("SELECT 1"))
                await eng.dispose()  # type: ignore[attr-defined]

            asyncio.run(_ping_and_dispose())
        else:
            with eng.begin() as conn:  # type: ignore[call-arg]
                conn.execute(text("SELECT 1"))
            eng.dispose()  # type: ignore[attr-defined]
    except OperationalError as exc:  # pragma: no cover (depends on env)
        raise RuntimeError(f"Failed to connect to database: {exc}") from exc


def repair_alembic_state_if_needed(cfg: Config) -> None:
    """If DB points to a non-existent local revision, reset to base (drop alembic_version)."""
    db_url = cfg.get_main_option("sqlalchemy.url") or os.getenv("SQL_URL")
    if not db_url:
        return

    # Gather local revision ids from versions/
    script_location = Path(cfg.get_main_option("script_location"))
    versions_dir = script_location / "versions"
    local_ids: Set[str] = set()
    if versions_dir.exists():
        for p in versions_dir.glob("*.py"):
            try:
                txt = p.read_text(encoding="utf-8")
            except Exception:
                continue
            for line in txt.splitlines():
                line = line.strip()
                if line.startswith("revision ="):
                    rid = line.split("=", 1)[1].strip().strip("'\"")
                    local_ids.add(rid)
                    break

    url_obj = make_url(db_url)
    if is_async_url(url_obj):

        async def _run() -> None:
            eng = build_engine(url_obj)  # AsyncEngine
            try:
                async with eng.begin() as conn:
                    # Do sync-y inspector / SQL via run_sync
                    def _check_and_maybe_drop(sync_conn):
                        insp = inspect(sync_conn)
                        if not insp.has_table("alembic_version"):
                            return
                        rows = list(
                            sync_conn.execute(
                                text("SELECT version_num FROM alembic_version")
                            ).fetchall()
                        )
                        missing = any((ver not in local_ids) for (ver,) in rows)
                        if missing:
                            sync_conn.execute(text("DROP TABLE IF EXISTS alembic_version"))

                    await conn.run_sync(_check_and_maybe_drop)
            finally:
                await eng.dispose()

        asyncio.run(_run())
    else:
        eng = build_engine(url_obj)  # sync Engine
        try:
            with eng.begin() as c:
                insp = inspect(c)
                if not insp.has_table("alembic_version"):
                    return
                rows = list(c.execute(text("SELECT version_num FROM alembic_version")).fetchall())
                missing = any((ver not in local_ids) for (ver,) in rows)
                if missing:
                    c.execute(text("DROP TABLE IF EXISTS alembic_version"))
        finally:
            eng.dispose()


def render_env_py(packages: Sequence[str], *, async_db: bool | None = None) -> str:
    """Render Alembic env.py content from packaged templates.

    - If async_db is None, detect from SQL_URL; default to sync if unknown.
    """
    import importlib.resources as pkg

    from sqlalchemy.engine import make_url as _make_url

    if async_db is None:
        try:
            db_url = get_database_url_from_env(required=False)
            async_db = bool(db_url and is_async_url(_make_url(db_url)))
        except Exception:
            async_db = False

    pkg_list = ", ".join(repr(p) for p in packages)
    tmpl_root = pkg.files("svc_infra.db.sql.templates.setup")
    name = "env_async.py.tmpl" if async_db else "env_sync.py.tmpl"
    txt = tmpl_root.joinpath(name).read_text(encoding="utf-8")
    return txt.replace("__PACKAGES_LIST__", pkg_list)


def build_alembic_config(
    project_root: Path | str, *, script_location: str = "migrations"
) -> Config:
    from alembic.config import Config as _Config

    root = Path(project_root).resolve()
    cfg_path = root / "alembic.ini"
    cfg = _Config(str(cfg_path)) if cfg_path.exists() else _Config()
    cfg.set_main_option("script_location", str((root / script_location).resolve()))

    env_db_url = os.getenv("SQL_URL", "").strip()
    if env_db_url:
        u = make_url(env_db_url)
        u = _ensure_ssl_default(u)
        if not is_async_url(u):
            u = _coerce_sync_driver(u)
        cfg.set_main_option("sqlalchemy.url", u.render_as_string(hide_password=False))

    # >>> ADD THIS GUARD <<<
    if not cfg.get_main_option("sqlalchemy.url"):
        raise RuntimeError(
            "No SQLAlchemy URL resolved. Pass `database_url` to the calling function "
            "or set SQL_URL in the environment."
        )

    cfg.set_main_option("path_separator", "os")
    cfg.set_main_option("prepend_sys_path", str(root))
    return cfg


def ensure_db_at_head(cfg: Config) -> None:
    """Idempotently bring the database to head; safe if already current."""
    from alembic import command as _command

    _command.upgrade(cfg, "head")


__all__ = [
    # env helpers
    "get_database_url_from_env",
    "is_async_url",
    "with_database",
    # engines and db bootstrap
    "build_engine",
    "ensure_database_exists",
    # setup helpers
    "render_env_py",
    "build_alembic_config",
    "ensure_db_at_head",
]



from __future__ import annotations

import os
from pathlib import Path
from typing import Optional, Sequence

from dotenv import load_dotenv

from .constants import DEFAULT_MONGO_DB_ENV_VARS, DEFAULT_MONGO_ENV_VARS


def prepare_process_env(project_root: Path | str) -> Path:
    """
    Prepare process environment (dotenv + PYTHONPATH) for code generation/tasks.
    Mirrors the SQL helper but does not touch SQL_URL.
    """
    root = Path(project_root).resolve()
    load_dotenv(dotenv_path=root / ".env", override=False)
    os.environ.setdefault("SKIP_APP_INIT", "1")

    # Make <project>/src importable (src-layout projects)
    src_dir = root / "src"
    if src_dir.exists():
        sys_path = os.environ.get("PYTHONPATH", "")
        parts = [str(src_dir)] + ([sys_path] if sys_path else [])
        os.environ["PYTHONPATH"] = os.pathsep.join(parts)
    return root


def _read_secret_from_file(path: str) -> Optional[str]:
    try:
        p = Path(path)
        if p.exists():
            return p.read_text(encoding="utf-8").strip()
    except Exception:
        pass
    return None


def get_mongo_url_from_env(
    required: bool = True,
    env_vars: Sequence[str] = DEFAULT_MONGO_ENV_VARS,
) -> Optional[str]:
    """
    Resolve the Mongo connection string with support for:
      - Primary env vars (DEFAULT_MONGO_ENV_VARS).
      - Companion *_FILE secret envs.
      - Conventional secret file path env MONGO_URL_FILE.
      - Default docker mounts: /run/secrets/mongo_url
    When found, writes it into os.environ["MONGO_URL"].
    """
    load_dotenv(override=False)

    # 1) direct envs (and allow absolute file pointers)
    for key in env_vars:
        val = os.getenv(key)
        if val and val.strip():
            s = val.strip()
            if s.startswith("file:"):
                s = s[5:]
            if os.path.isabs(s) and Path(s).exists():
                file_val = _read_secret_from_file(s)
                if file_val:
                    os.environ["MONGO_URL"] = file_val
                    return file_val
            os.environ["MONGO_URL"] = s
            return s

        file_key = f"{key}_FILE"
        file_path = os.getenv(file_key)
        if file_path:
            file_val = _read_secret_from_file(file_path)
            if file_val:
                os.environ["MONGO_URL"] = file_val
                return file_val

    # 2) conventional secret env
    fp = os.getenv("MONGO_URL_FILE")
    if fp:
        file_val = _read_secret_from_file(fp)
        if file_val:
            os.environ["MONGO_URL"] = file_val
            return file_val

    # 3) docker/k8s default secret mount
    file_val = _read_secret_from_file("/run/secrets/mongo_url")
    if file_val:
        os.environ["MONGO_URL"] = file_val
        return file_val

    if required:
        raise RuntimeError(
            "Mongo URL not set. Set MONGO_URL (or MONGODB_URL) or provide *_FILE secret."
        )
    return None


def get_mongo_dbname_from_env(
    required: bool = False,
    env_vars: Sequence[str] = DEFAULT_MONGO_DB_ENV_VARS,
    default: str = "app",
) -> Optional[str]:
    """Return a database name from env; optional (Motor can connect without it)."""
    load_dotenv(override=False)
    for key in env_vars:
        val = os.getenv(key)
        if val and val.strip():
            os.environ["MONGO_DB"] = val.strip()
            return val.strip()
    if required:
        raise RuntimeError("Mongo DB name not set. Set MONGO_DB.")
    return os.getenv("MONGO_DB") or None



from __future__ import annotations

import importlib
from typing import Iterable, Optional


class _MissingPrometheus(Exception):
    pass


def _require_prometheus() -> None:
    try:
        importlib.import_module("prometheus_client")
    except Exception as exc:  # pragma: no cover - optional dep path
        raise ImportError(
            "Prometheus metrics require prometheus-client. Install via 'pip install svc-infra[metrics]'."
        ) from exc


def _prom_mod():
    _require_prometheus()
    return importlib.import_module("prometheus_client")


def registry():
    """Return the current Prometheus registry, handling multiprocess mode if set.

    Uses the default global registry unless PROMETHEUS_MULTIPROC_DIR is set,
    in which case attaches a MultiProcessCollector to a fresh CollectorRegistry.
    """
    import os

    prom = _prom_mod()
    REGISTRY = getattr(prom, "REGISTRY")
    CollectorRegistry = getattr(prom, "CollectorRegistry")
    multiprocess = getattr(prom, "multiprocess", None)

    if os.environ.get("PROMETHEUS_MULTIPROC_DIR") and multiprocess is not None:
        reg = CollectorRegistry()
        multiprocess.MultiProcessCollector(reg)
        return reg
    return REGISTRY


def _mk_metric(
    ctor_name: str,
    name: str,
    doc: str,
    labels: Optional[Iterable[str]] = None,
    **kwargs,
):
    prom = _prom_mod()
    Counter = getattr(prom, "Counter")
    Gauge = getattr(prom, "Gauge")
    Histogram = getattr(prom, "Histogram")
    Summary = getattr(prom, "Summary")

    ctors = {
        "Counter": Counter,
        "Gauge": Gauge,
        "Histogram": Histogram,
        "Summary": Summary,
    }
    ctor = ctors[ctor_name]
    labelnames = list(labels) if labels else None
    metric = ctor(name, doc, labelnames=labelnames, **kwargs)
    return metric


def counter(name: str, doc: str, labels: Optional[Iterable[str]] = None):
    return _mk_metric("Counter", name, doc, labels)


def gauge(name: str, doc: str, labels: Optional[Iterable[str]] = None, **kw):
    # e.g. gauge(..., multiprocess_mode="livesum")
    return _mk_metric("Gauge", name, doc, labels, **kw)


def histogram(
    name: str,
    doc: str,
    labels: Optional[Iterable[str]] = None,
    buckets: Optional[Iterable[float]] = None,
):
    kwargs = {"buckets": list(buckets) if buckets else None}
    # Remove None so prometheus-client uses its defaults
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    return _mk_metric("Histogram", name, doc, labels, **kwargs)


def summary(name: str, doc: str, labels: Optional[Iterable[str]] = None):
    return _mk_metric("Summary", name, doc, labels)



from __future__ import annotations

import logging
from typing import Iterable


def _is_metrics_like(record: logging.LogRecord, paths: Iterable[str]) -> bool:
    """
    Heuristics to detect access-log lines for given paths across uvicorn/gunicorn.
    Works with both dict- and string-formatted records.
    """
    try:
        # uvicorn/gunicorn commonly put 'request_line' in args (dict)
        if isinstance(record.args, dict):
            rl = record.args.get("request_line") or record.args.get("r")
            if rl and isinstance(rl, str):
                # request line looks like: "GET /metrics HTTP/1.1"
                for p in paths:
                    if f" {p} " in f" {rl} ":
                        return True
        # Fallback: substring match in rendered message
        msg = record.getMessage()
        if isinstance(msg, str):
            for p in paths:
                if f" {p} " in f" {msg} ":
                    return True
    except Exception:
        pass
    return False


class _DropPathsFilter(logging.Filter):
    def __init__(self, paths: Iterable[str]):
        super().__init__()
        self._paths = tuple(paths)

    def filter(self, record: logging.LogRecord) -> bool:
        # Return False to DROP the record.
        return not _is_metrics_like(record, self._paths)


def filter_logs_for_paths(
    *,
    paths: Iterable[str] = ("/metrics",),
    enabled: bool = True,
) -> None:
    """
    When enabled=True, suppress access logs for given paths on common web servers.
    Safe to call multiple times.
    """
    if not enabled:
        return

    filt = _DropPathsFilter(paths)

    # Common access loggers to target
    for name in ("uvicorn.access", "gunicorn.access"):
        try:
            logger = logging.getLogger(name)
            # Avoid stacking duplicates
            if not any(isinstance(f, _DropPathsFilter) for f in logger.filters):
                logger.addFilter(filt)
        except Exception:
            pass



from __future__ import annotations

from typing import Sequence

# Environment variable names to look up for Mongo URL
DEFAULT_MONGO_ENV_VARS: Sequence[str] = (
    "MONGO_URL",
    "MONGODB_URL",
)

# Default DB name envs (optional convenience)
DEFAULT_MONGO_DB_ENV_VARS: Sequence[str] = (
    "MONGO_DB",
    "MONGODB_DB",
    "MONGO_DATABASE",
)

__all__ = ["DEFAULT_MONGO_ENV_VARS", "DEFAULT_MONGO_DB_ENV_VARS"]



from __future__ import annotations

import os
import subprocess
from pathlib import Path
from typing import Iterable, Optional

DEFAULT_SENTRIES: tuple[str, ...] = (
    ".git",
    ".hg",
    ".svn",
    "alembic.ini",
    "migrations",
    "migrations/env.py",
    "migrations/script.py.mako",
    "pyproject.toml",
    "setup.cfg",
    "decorators.py",
    "Pipfile",
    "requirements.txt",
    "poetry.lock",
    "pdm.lock",
    "uv.lock",
    "hatch.toml",
    "src",
    "app",
    "backend",
    ".project-root",
    ".svc-infra-root",
)

ENV_VAR = "PROJECT_ROOT"


def _is_root_marker(dir_: Path, sentries: Iterable[str]) -> bool:
    return any((dir_ / name).exists() for name in sentries)


def _git_toplevel(start: Path) -> Optional[Path]:
    try:
        out = subprocess.check_output(
            ["git", "rev-parse", "--show-toplevel"],
            cwd=str(start),
            stderr=subprocess.DEVNULL,
        )
        p = Path(out.decode().strip())
        return p if p.exists() else None
    except Exception:
        return None


def resolve_project_root(
    start: Optional[Path] = None,
    *,
    env_var: str = ENV_VAR,
    extra_sentries: Iterable[str] = (),
) -> Path:
    env = os.getenv(env_var)
    if env:
        p = Path(env).expanduser().resolve()
        if p.is_dir():
            return p

    start = (start or Path.cwd()).resolve()

    git_root = _git_toplevel(start)
    if git_root:
        return git_root

    sentries = tuple(DEFAULT_SENTRIES) + tuple(extra_sentries)
    for d in [start, *start.parents]:
        if _is_root_marker(d, sentries):
            return d

    return start



from __future__ import annotations

import asyncio
import shutil
from pathlib import Path
from typing import List, Optional


def _has_poetry(root: Path) -> bool:
    return (root / "pyproject.toml").exists() and bool(shutil.which("poetry"))


def candidate_cmds(root: Path, prog: str, argv: List[str]) -> List[List[str]]:
    """
    Return argv lists to try in order:
      1) poetry run <prog> ...
      2) <prog> ...
      3) python -m <module> ...
    """
    cmds: List[List[str]] = []
    if _has_poetry(root):
        cmds.append(["poetry", "run", prog, *argv])

    if shutil.which(prog):
        cmds.append([prog, *argv])

    py = shutil.which("python3") or shutil.which("python") or "python"
    module = prog.replace("-", "_") + ".cli_shim"  # e.g., svc-infra -> svc_infra.cli_shim
    cmds.append([py, "-m", module, *argv])

    return cmds


async def run_from_root(root: Path, prog: str, argv: List[str]) -> str:
    """
    cd to project root and run the first working candidate command.
    Returns captured stdout+stderr text; raises on total failure.
    """
    last_exc: Optional[BaseException] = None
    for cmd in candidate_cmds(root, prog, argv):
        try:
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                cwd=str(root),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
            out, _ = await proc.communicate()
            if proc.returncode == 0:
                return out.decode(errors="replace")
            last_exc = RuntimeError(
                f"Exit {proc.returncode}: {' '.join(cmd)}\n{out.decode(errors='replace')}"
            )
        except Exception as e:
            last_exc = e
            continue
    raise RuntimeError(f"All runners failed in {root} for: {prog} {' '.join(argv)}") from last_exc



from __future__ import annotations

import importlib
import logging
import pkgutil
from types import ModuleType
from typing import Optional

from fastapi import FastAPI
from fastapi.routing import APIRoute

from svc_infra.app.env import ALL_ENVIRONMENTS, CURRENT_ENVIRONMENT, DEV_ENV, LOCAL_ENV, Environment

logger = logging.getLogger(__name__)


def _should_skip_module(module_name: str) -> bool:
    """
    Returns True if the module should be skipped based on:
    - private/dunder final segment
    """
    parts = module_name.split(".")
    last_segment = parts[-1]
    return last_segment.startswith("_")


def _derive_docs_from_module(module: ModuleType) -> tuple[str | None, str | None]:
    # 1) explicit constants win
    mod_summary = getattr(module, "ROUTER_SUMMARY", None)
    mod_description = getattr(module, "ROUTER_DESCRIPTION", None)
    if mod_summary or mod_description:
        return mod_summary, mod_description

    # 2) fallback: module docstring (first non-empty line = summary; rest = description)
    doc = (module.__doc__ or "").strip()
    if not doc:
        return None, None
    lines = [ln.strip() for ln in doc.splitlines()]
    lines = [ln for ln in lines if ln]  # drop empties
    if not lines:
        return None, None
    summary = lines[0]
    description = "\n".join(lines[1:]) if len(lines) > 1 else None
    return summary, description


def _validate_base_package(base_package: str) -> ModuleType:
    """Validate and import the base package."""
    try:
        package_module: ModuleType = importlib.import_module(base_package)
    except Exception as exc:
        raise RuntimeError(f"Could not import base_package '{base_package}': {exc}") from exc

    if not hasattr(package_module, "__path__"):
        raise RuntimeError(
            f"Provided base_package '{base_package}' is not a package (no __path__)."
        )

    return package_module


def _normalize_environment(environment: Optional[Environment | str]) -> Environment:
    """Normalize the environment parameter."""
    return (
        CURRENT_ENVIRONMENT
        if environment is None
        else (Environment(environment) if not isinstance(environment, Environment) else environment)
    )


def _should_force_include_in_schema(
    environment: Environment, force_include_in_schema: Optional[bool]
) -> bool:
    """Determine if routers should be forced to include in schema."""
    if force_include_in_schema is None:
        return environment in (LOCAL_ENV, DEV_ENV)
    return force_include_in_schema


def _is_router_excluded_by_environment(
    module: ModuleType, environment: Environment, module_name: str
) -> bool:
    """Check if router should be excluded based on environment restrictions."""
    router_excluded_envs = getattr(module, "ROUTER_EXCLUDED_ENVIRONMENTS", None)
    if router_excluded_envs is None:
        return False

    # Support ALL_ENVIRONMENTS as a special value
    if router_excluded_envs is ALL_ENVIRONMENTS or (
        isinstance(router_excluded_envs, set) and router_excluded_envs == ALL_ENVIRONMENTS
    ):
        logger.debug(f"Skipping router module {module_name} due to ALL_ENVIRONMENTS exclusion.")
        return True

    # Normalize to set of Environment or str
    if not isinstance(router_excluded_envs, (set, list, tuple)):
        logger.warning(
            f"ROUTER_EXCLUDED_ENVIRONMENTS in {module_name} must be a set/list/tuple, got {type(router_excluded_envs)}"
        )
        return False

    normalized_excluded_envs = set()
    for e in router_excluded_envs:
        try:
            normalized_excluded_envs.add(Environment(e) if not isinstance(e, Environment) else e)
        except Exception:
            normalized_excluded_envs.add(str(e))

    if environment in normalized_excluded_envs or str(environment) in normalized_excluded_envs:
        logger.debug(
            f"Skipping router module {module_name} due to ROUTER_EXCLUDED_ENVIRONMENTS restriction: {router_excluded_envs}"
        )
        return True

    return False


def _is_router_included_by_environment(
    module: ModuleType, environment: Environment, module_name: str
) -> bool:
    router_envs = getattr(module, "ROUTER_ENVIRONMENTS", None)
    if router_envs is None:
        return True
    if not isinstance(router_envs, (set, list, tuple)):
        logger.warning(
            f"ROUTER_ENVIRONMENTS in {module_name} must be a set/list/tuple, got {type(router_envs)}"
        )
        return True
    normalized = set()
    for e in router_envs:
        try:
            normalized.add(Environment(e) if not isinstance(e, Environment) else e)
        except Exception:
            normalized.add(str(e))
    inc = environment in normalized or str(environment) in normalized
    if not inc:
        logger.debug(
            f"Skipping router module {module_name} due to ROUTER_ENVIRONMENTS restriction: {router_envs}"
        )
    return inc


def _should_never_include_in_schema(module: ModuleType) -> bool:
    """Check if router should never be included in schema."""
    return getattr(module, "ROUTER_NEVER_IN_SCHEMA", False) is True


def _apply_default_docs_to_routes(
    router, default_summary: str | None, default_description: str | None
) -> None:
    """Apply default summary and description to routes that don't have them."""
    for r in getattr(router, "routes", []):
        if isinstance(r, APIRoute):
            if not r.summary and default_summary:
                r.summary = default_summary
            if not r.description and default_description:
                r.description = default_description


def _build_include_kwargs(module: ModuleType, prefix: str, force_include: bool) -> dict:
    """Build the kwargs for app.include_router."""
    router_prefix = getattr(module, "ROUTER_PREFIX", None)
    router_tag = getattr(module, "ROUTER_TAG", None)
    include_in_schema = getattr(module, "INCLUDE_ROUTER_IN_SCHEMA", True)

    include_kwargs = {"prefix": prefix}
    if router_prefix:
        include_kwargs["prefix"] = prefix.rstrip("/") + router_prefix
    if router_tag:
        include_kwargs["tags"] = [router_tag]

    # the key line: force in LOCAL, otherwise respect the module
    include_kwargs["include_in_schema"] = True if force_include else include_in_schema

    return include_kwargs


def _process_router_module(
    app: FastAPI,
    module: ModuleType,
    module_name: str,
    prefix: str,
    environment: Environment,
    force_include: bool,
) -> bool:
    """Process a single router module. Returns True if router was included, False if skipped."""
    if not _is_router_included_by_environment(module, environment, module_name):
        return False

    router = getattr(module, "router", None)
    if router is None:
        return False

    # Check for environment exclusions
    if _is_router_excluded_by_environment(module, environment, module_name):
        return False

    # Check for never-in-schema opt-out
    if _should_never_include_in_schema(module):
        return False

    # Derive default docs for this module
    default_summary, default_description = _derive_docs_from_module(module)

    # Pre-fill missing per-route summaries/descriptions
    _apply_default_docs_to_routes(router, default_summary, default_description)

    # Build include kwargs
    include_kwargs = _build_include_kwargs(module, prefix, force_include)

    # Include the router
    app.include_router(router, **include_kwargs)
    logger.debug(
        "Included router from module: %s (prefix=%s, tag=%s, include_in_schema=%s)",
        module_name,
        include_kwargs.get("prefix"),
        include_kwargs.get("tags", [None])[0] if include_kwargs.get("tags") else None,
        include_kwargs.get("include_in_schema"),
    )

    return True


def register_all_routers(
    app: FastAPI,
    *,
    base_package: Optional[str] = None,
    prefix: str = "",
    environment: Optional[Environment | str] = None,
    force_include_in_schema: Optional[bool] = None,
) -> None:
    """
    Recursively discover and register all FastAPI routers under a routers package.

    Args:
        app: FastAPI application instance.
        base_package: Import path to the root routers package (e.g., "myapp.api.routers").
            If omitted, derived from this module's package.
        prefix: API prefix for all routers (e.g., "/v0").
        environment: The current environment (defaults to get_env()).

    Behavior:
        - Any module under the package with a top-level `router` variable is included.
        - Files/packages whose final segment starts with '_' are skipped.
        - If a module defines ROUTER_ENVIRONMENTS, it is a set/list of environments (Env or str) in which the router is included.
        - Import errors are logged and skipped.
        - Nested discovery requires `__init__.py` files in packages.
        - If a module defines ROUTER_PREFIX or ROUTER_TAGS, they are used for that router.
    """
    if base_package is None:
        if __package__ is None:
            raise RuntimeError("Cannot derive base_package; please pass base_package explicitly.")
        base_package = __package__

    package_module = _validate_base_package(base_package)
    environment = _normalize_environment(environment)
    force_include = _should_force_include_in_schema(environment, force_include_in_schema)

    for _, module_name, _ in pkgutil.walk_packages(
        package_module.__path__, prefix=f"{base_package}."
    ):
        if _should_skip_module(module_name):
            logger.debug("Skipping router module due to exclusion/private: %s", module_name)
            continue

        try:
            module = importlib.import_module(module_name)
        except Exception as exc:
            logger.exception("Failed to import router module %s: %s", module_name, exc)
            continue

        _process_router_module(app, module, module_name, prefix, environment, force_include)



from __future__ import annotations

from typing import Any, Optional, Type, Union, get_args, get_origin

from bson import ObjectId
from pydantic import BaseModel

from svc_infra.db.crud_schema import FieldSpec, make_crud_schemas_from_specs
from svc_infra.db.nosql.types import PyObjectId

_MUTABLE_ID_FIELDS = {"id", "_id"}
_TS_FIELDS = {"created_at", "updated_at"}


def _is_optional(annotation: Any) -> bool:
    return get_origin(annotation) is Optional or (
        get_origin(annotation) is Union and type(None) in get_args(annotation)
    )


def _unwrap_union(annotation: Any) -> set[type]:
    """
    Return the set of non-None types inside a Union/Optional[T] or {annotation} if not a Union.
    """
    origin = get_origin(annotation)
    if origin is Union:
        return {t for t in get_args(annotation) if t is not type(None)}  # noqa: E721
    return {annotation} if annotation is not None else set()


def _is_objectid_like(annotation: Any) -> bool:
    """
    True if the annotation is (or wraps) bson.ObjectId / PyObjectId.
    """
    inner = _unwrap_union(annotation)
    return any(t in (ObjectId, PyObjectId) for t in inner)


def make_document_crud_schemas(
    document_model: Type[BaseModel],
    *,
    create_exclude: tuple[str, ...] = ("_id",),
    read_name: str | None = None,
    create_name: str | None = None,
    update_name: str | None = None,
    read_exclude: tuple[str, ...] = (),
    update_exclude: tuple[str, ...] = (),
    json_encoders: dict[type, Any] | None = None,
) -> tuple[type[BaseModel], type[BaseModel], type[BaseModel]]:
    """
    Derive (Read, Create, Update) from a Pydantic document model (Mongo).

    Key behavior:
    - Read: expose fields, but coerce any ObjectId-like types to `str` so responses are JSON-safe.
    - Create: exclude ID/timestamps and anything in create_exclude; required mirrors document model.
    - Update: all optional; ObjectId-like types coerced to `str` for JSON safety.
    """
    annotations = document_model.model_fields  # Pydantic v2
    explicit_create_ex = set(create_exclude) | _MUTABLE_ID_FIELDS | _TS_FIELDS
    read_ex = set(read_exclude)
    update_ex = set(update_exclude)

    specs: list[FieldSpec] = []
    for name, field in annotations.items():
        T = field.annotation or Any
        required = field.is_required()

        # Coerce ObjectId-ish to `str` for READ/UPDATE to avoid custom encoder headaches.
        # (Create path ignores IDs anyway due to explicit_create_ex.)
        T_out = str if _is_objectid_like(T) else T

        specs.append(
            FieldSpec(
                name=name,
                typ=T_out,
                required_for_create=bool(
                    required and not _is_optional(T) and name not in explicit_create_ex
                ),
                exclude_from_create=(name in explicit_create_ex),
                exclude_from_read=(name in read_ex),
                exclude_from_update=(name in update_ex),
            )
        )

    # Backstop encoders in case any exotic types slip through
    encoders = {ObjectId: str, PyObjectId: str}
    if json_encoders:
        encoders.update(json_encoders)

    return make_crud_schemas_from_specs(
        specs=specs,
        read_name=read_name or f"{document_model.__name__}Read",
        create_name=create_name or f"{document_model.__name__}Create",
        update_name=update_name or f"{document_model.__name__}Update",
        json_encoders=encoders,
    )



from __future__ import annotations

import json
import os
import re
import urllib.request
from importlib.resources import files

# ---------------- helpers ----------------


def _gapi(base, token, path, method="GET", body=None):
    url = f"{base.rstrip('/')}{path}"
    data = json.dumps(body).encode("utf-8") if body is not None else None
    req = urllib.request.Request(
        url,
        data=data,
        method=method,
        headers={
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )
    with urllib.request.urlopen(req) as r:
        ct = r.headers.get("Content-Type", "")
        raw = r.read()
        return json.loads(raw.decode("utf-8")) if "application/json" in ct else None


def _find_prom_uid(base, token):
    dss = _gapi(base, token, "/api/datasources") or []
    # prefer any grafana cloud/mimir url if present
    for ds in dss:
        if ds.get("type") == "prometheus" and "grafana.net" in (ds.get("url") or ""):
            return ds["uid"]
    for ds in dss:
        if ds.get("type") == "prometheus":
            return ds["uid"]
    raise RuntimeError("No Prometheus datasource found in Grafana Cloud")


def _ensure_folder(base, token, title):
    for f in _gapi(base, token, "/api/folders") or []:
        if f.get("title") == title:
            return f
    return _gapi(base, token, "/api/folders", method="POST", body={"title": title})


def _slug(s: str) -> str:
    s = s.strip().lower()
    s = re.sub(r"[^a-z0-9._-]+", "-", s)
    s = re.sub(r"-+", "-", s).strip("-")
    return s or "dashboard"


def _stable_uid(name: str) -> str:
    base = f"svcinfra-{_slug(name)}"
    return base[:40]


def _patch_ds(d, prom_uid):
    ds = {"type": "prometheus", "uid": prom_uid}
    dd = json.loads(json.dumps(d))  # deep copy
    dd["datasource"] = ds
    for p in dd.get("panels", []) or []:
        p["datasource"] = ds
        for t in p.get("targets", []) or []:
            t["datasource"] = ds
        for sp in p.get("panels", []) or []:
            sp["datasource"] = ds
            for t in sp.get("targets", []) or []:
                t["datasource"] = ds
    templ = (dd.get("templating") or {}).get("list") or []
    for v in templ:
        v["datasource"] = ds
        if v.get("includeAll"):
            v["allValue"] = ".*"
        v.setdefault("refresh", 2)
    return dd


def _rewrite_rate_windows(d: dict) -> dict:
    """Optionally replace $__rate_interval with a fixed window from env."""
    win = os.getenv("SVC_INFRA_RATE_WINDOW", "").strip()
    if not win:
        return d

    dd = json.loads(json.dumps(d))
    for p in dd.get("panels", []) or []:
        targets = p.get("targets") or []
        for t in targets:
            expr = t.get("expr")
            if isinstance(expr, str) and "$__rate_interval" in expr:
                t["expr"] = expr.replace("$__rate_interval", win)
    return dd


def push_dashboards_from_pkg(
    base_url: str, token: str, folder_title: str = "Service Infrastructure"
):
    prom_uid = _find_prom_uid(base_url, token)
    folder = _ensure_folder(base_url, token, folder_title)
    fuid = folder["uid"]

    # env-driven knobs
    refresh = os.getenv("SVC_INFRA_DASHBOARD_REFRESH", "5s")
    default_from = os.getenv("SVC_INFRA_DASHBOARD_RANGE", "now-6h")
    default_to = "now"

    dash_root = files("svc_infra.obs.providers.grafana").joinpath("dashboards")
    for res in dash_root.iterdir():
        if not res.name.endswith(".json"):
            continue

        d = json.loads(res.read_text(encoding="utf-8"))
        d.setdefault("title", res.name[:-5])
        d.setdefault("uid", _stable_uid(d["title"]))
        d["id"] = None  # upsert by uid

        # apply datasource/templating patch
        d = _patch_ds(d, prom_uid)
        # apply rate window rewrite if requested
        d = _rewrite_rate_windows(d)

        # env-driven dashboard refresh + default time range
        d["refresh"] = refresh
        d["time"] = {"from": default_from, "to": default_to}

        payload = {
            "dashboard": d,
            "folderUid": fuid,
            "overwrite": True,
            "message": "svc-infra auto-sync",
        }
        _gapi(base_url, token, "/api/dashboards/db", method="POST", body=payload)



from __future__ import annotations

from pydantic import BaseModel
from sqlalchemy import Column
from sqlalchemy.orm import Mapper, class_mapper

from svc_infra.db.crud_schema import FieldSpec, make_crud_schemas_from_specs


def _sa_columns(model: type[object]) -> list[Column]:
    mapper: Mapper = class_mapper(model)  # raises if not a mapped class
    return list(mapper.columns)


def _py_type(col: Column) -> type:
    # Prefer SQLAlchemy-provided python_type when available
    if getattr(col.type, "python_type", None):
        return col.type.python_type  # type: ignore[no-any-return]

    from datetime import date, datetime
    from typing import Any as _Any
    from uuid import UUID

    from sqlalchemy import JSON, Boolean, Date, DateTime, Integer, String, Text

    try:
        from sqlalchemy.dialects.postgresql import JSONB
        from sqlalchemy.dialects.postgresql import UUID as PG_UUID  # type: ignore
    except Exception:  # pragma: no cover
        PG_UUID = None  # type: ignore
        JSONB = None  # type: ignore

    t = col.type
    if PG_UUID is not None and isinstance(t, PG_UUID):
        return UUID
    if isinstance(t, (String, Text)):
        return str
    if isinstance(t, Integer):
        return int
    if isinstance(t, Boolean):
        return bool
    if isinstance(t, (DateTime,)):
        return datetime
    if isinstance(t, (Date,)):
        return date
    if isinstance(t, JSON):
        return dict
    if JSONB is not None and isinstance(t, JSONB):
        return dict
    return _Any


def _exclude_from_create(col: Column) -> bool:
    """Heuristics for excluding columns from Create schema."""
    if col.primary_key:
        return True
    if col.server_default is not None:
        return True
    default = getattr(col, "default", None)
    if getattr(default, "is_sequence", False):
        return True
    if getattr(default, "arg", None):  # e.g., default=uuid.uuid4
        return True
    if col.onupdate is not None:
        return True
    if col.name in {"created_at", "updated_at"}:
        return True
    return False


def make_crud_schemas(
    model: type[object],
    *,
    create_exclude: tuple[str, ...] = ("id",),
    read_name: str | None = None,
    create_name: str | None = None,
    update_name: str | None = None,
    read_exclude: tuple[str, ...] = (),
    update_exclude: tuple[str, ...] = (),
) -> tuple[type[BaseModel], type[BaseModel], type[BaseModel]]:
    cols = _sa_columns(model)

    specs: list[FieldSpec] = []
    explicit_excludes = set(create_exclude)
    read_ex = set(read_exclude)
    update_ex = set(update_exclude)

    for col in cols:
        name = col.name
        T = _py_type(col)

        is_required = (
            not col.nullable
            and col.default is None
            and col.server_default is None
            and not col.primary_key
        )

        specs.append(
            FieldSpec(
                name=name,
                typ=T,
                required_for_create=bool(
                    is_required and name not in explicit_excludes and not _exclude_from_create(col)
                ),
                exclude_from_create=bool(name in explicit_excludes or _exclude_from_create(col)),
                exclude_from_read=bool(name in read_ex),
                exclude_from_update=bool(name in update_ex),
            )
        )

    return make_crud_schemas_from_specs(
        specs=specs,
        read_name=read_name or f"{model.__name__}Read",
        create_name=create_name or f"{model.__name__}Create",
        update_name=update_name or f"{model.__name__}Update",
    )



from __future__ import annotations

import logging
from typing import Optional

from cashews import cache as _cache

# Module-level logger
logger = logging.getLogger(__name__)

# Configuration constants
DEFAULT_PREFIX = "svc"
DEFAULT_VERSION = "v1"
DEFAULT_READINESS_TIMEOUT = 5.0
PROBE_KEY_SUFFIX = "__probe__"
PROBE_VALUE = "ok"
PROBE_EXPIRE_SECONDS = 3

# Global state
_current_prefix: str = DEFAULT_PREFIX
_current_version: str = DEFAULT_VERSION


def alias() -> str:
    """
    Get the human-readable namespace label.

    Returns:
        Formatted namespace string like "svc:v1"
    """
    return f"{_current_prefix}:{_current_version}"


def _full_prefix() -> str:
    """
    Get the actual key prefix applied by cashews.

    Returns:
        Formatted prefix string with trailing colon like "svc:v1:"
    """
    return f"{_current_prefix}:{_current_version}:"


def setup_cache(
    url: Optional[str] = None,
    *,
    prefix: Optional[str] = None,
    version: Optional[str] = None,
):
    """
    Configure Cashews and set a global key prefix for namespacing.

    Args:
        url: Cache backend URL (Redis, etc.). If None, uses default configuration
        prefix: Cache key prefix override. Defaults to "svc"
        version: Cache version override. Defaults to "v1"

    Returns:
        Awaitable setup result from cashews that callers may await or not

    Example:
        Basic setup:
        >>> setup_cache()

        With custom Redis URL:
        >>> setup_cache(url="redis://localhost:6379")

        With custom namespace:
        >>> setup_cache(prefix="myapp", version="v2")
    """
    global _current_prefix, _current_version

    # Update global state if new values provided
    if prefix is not None:
        _current_prefix = prefix
        logger.info(f"Cache prefix updated to: {_current_prefix}")

    if version is not None:
        _current_version = version
        logger.info(f"Cache version updated to: {_current_version}")

    # Setup backend connection
    try:
        setup_awaitable = _cache.setup(url) if url else _cache.setup()
        logger.info(f"Cache backend setup initiated with URL: {url or 'default'}")
    except Exception as e:
        logger.error(f"Failed to setup cache backend: {e}")
        raise

    return setup_awaitable


async def wait_ready(timeout: float = DEFAULT_READINESS_TIMEOUT) -> None:
    """
    Wait for cache to be ready by performing a readiness probe.

    Args:
        timeout: Maximum time to wait for readiness (not currently used by probe)

    Raises:
        RuntimeError: If the readiness probe fails

    Example:
        >>> await wait_ready()
        >>> await wait_ready(timeout=10.0)
    """
    probe_key = f"{_full_prefix()}{PROBE_KEY_SUFFIX}"

    try:
        # Set probe value
        await _cache.set(probe_key, PROBE_VALUE, expire=PROBE_EXPIRE_SECONDS)
        logger.debug(f"Set readiness probe key: {probe_key}")

        # Verify probe value
        retrieved_value = await _cache.get(probe_key)

        if retrieved_value != PROBE_VALUE:
            error_msg = (
                f"Cache readiness probe failed. Expected '{PROBE_VALUE}', got '{retrieved_value}'"
            )
            logger.error(error_msg)
            raise RuntimeError(error_msg)

        logger.info("Cache readiness probe successful")

    except Exception as e:
        if isinstance(e, RuntimeError):
            raise
        error_msg = f"Cache readiness probe encountered error: {e}"
        logger.error(error_msg)
        raise RuntimeError(error_msg) from e


async def shutdown_cache() -> None:
    """
    Gracefully shutdown the cache backend.

    This method handles exceptions gracefully to ensure shutdown doesn't fail.
    """
    try:
        await _cache.close()
        logger.info("Cache backend shutdown successfully")
    except Exception as e:
        logger.warning(f"Error during cache shutdown (ignored): {e}")


def instance():
    """
    Get the underlying cashews cache instance.

    Returns:
        The cashews cache instance for direct access

    Warning:
        Direct access to the cache instance bypasses the namespace prefix.
        Use with caution in production code.
    """
    return _cache



from typing import Annotated, Any, Optional, Sequence, Type, cast

from fastapi import APIRouter, Body, Depends, HTTPException
from motor.motor_asyncio import AsyncIOMotorDatabase

from svc_infra.api.fastapi import public_router
from svc_infra.api.fastapi.db.http import (
    LimitOffsetParams,
    OrderParams,
    Page,
    SearchParams,
    dep_limit_offset,
    dep_order,
    dep_search,
)
from svc_infra.db.nosql.mongo.client import acquire_db
from svc_infra.db.nosql.service import NoSqlService

DBDep = Annotated[AsyncIOMotorDatabase, Depends(acquire_db)]


def _parse_sort(
    order_spec: Optional[str], allowed_order_fields: Optional[list[str]]
) -> list[tuple[str, int]]:
    if not order_spec:
        return []
    out: list[tuple[str, int]] = []
    for raw in [p.strip() for p in order_spec.split(",") if p.strip()]:
        field = raw[1:] if raw.startswith("-") else raw
        if allowed_order_fields and field not in allowed_order_fields:
            continue
        out.append((field, -1 if raw.startswith("-") else 1))
    return out


def make_crud_router_plus_mongo(
    *,
    service: NoSqlService,
    read_schema: Type[Any],
    create_schema: Type[Any],
    update_schema: Type[Any],
    prefix: str,
    tags: list[str] | None = None,
    search_fields: Optional[Sequence[str]] = None,
    default_ordering: Optional[str] = None,
    allowed_order_fields: Optional[list[str]] = None,
    mount_under_db_prefix: bool = True,
) -> APIRouter:
    router_prefix = ("/_mongo" + prefix) if mount_under_db_prefix else prefix
    router = public_router(
        prefix=router_prefix,
        tags=tags or [prefix.strip("/")],
        redirect_slashes=False,
    )

    # LIST
    @router.get("", response_model=cast(Any, Page[read_schema]))
    async def list_items(
        db: DBDep,
        lp: Annotated[LimitOffsetParams, Depends(dep_limit_offset)],
        op: Annotated[OrderParams, Depends(dep_order)],
        sp: Annotated[SearchParams, Depends(dep_search)],
    ):
        sort = _parse_sort(op.order_by or default_ordering, allowed_order_fields)
        if sp.q and search_fields:
            items = await service.search(
                db, q=sp.q, fields=search_fields, limit=lp.limit, offset=lp.offset, sort=sort
            )
            total = await service.count_filtered(db, q=sp.q, fields=search_fields)
        else:
            items = await service.list(db, limit=lp.limit, offset=lp.offset, sort=sort)
            total = await service.count(db)
        return Page[read_schema].from_items(
            total=total, items=items, limit=lp.limit, offset=lp.offset
        )

    # GET by id
    @router.get("/{item_id}", response_model=cast(Any, read_schema))
    async def get_item(db: DBDep, item_id: Any):
        row = await service.get(db, item_id)
        if not row:
            raise HTTPException(404, "Not found")
        return row

    # CREATE
    @router.post("", response_model=cast(Any, read_schema), status_code=201)
    async def create_item(db: DBDep, payload: create_schema = Body(...)):
        data = payload.model_dump(exclude_unset=True)
        return await service.create(db, data)

    # UPDATE
    @router.patch("/{item_id}", response_model=cast(Any, read_schema))
    async def update_item(db: DBDep, item_id: Any, payload: update_schema = Body(...)):
        data = payload.model_dump(exclude_unset=True)
        row = await service.update(db, item_id, data)
        if not row:
            raise HTTPException(404, "Not found")
        return row

    # DELETE
    @router.delete("/{item_id}", status_code=204)
    async def delete_item(db: DBDep, item_id: Any):
        ok = await service.delete(db, item_id)
        if not ok:
            raise HTTPException(404, "Not found")
        return

    return router



from __future__ import annotations

import smtplib
from email.message import EmailMessage
from typing import Protocol

from svc_infra.app.env import CURRENT_ENVIRONMENT, PROD_ENV

from .settings import get_auth_settings


class Sender(Protocol):
    def send(self, to: str, subject: str, html_body: str) -> None:
        pass


class ConsoleSender:
    def send(self, to: str, subject: str, html_body: str) -> None:
        print(f"[MAIL -> {to}] {subject}\n{html_body}\n")


class SMTPSender:
    def __init__(self, host: str, port: int, username: str, password: str, from_addr: str) -> None:
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.from_addr = from_addr

    def send(self, to: str, subject: str, html_body: str) -> None:
        msg = EmailMessage()
        msg["From"] = self.from_addr
        msg["To"] = to
        msg["Subject"] = subject
        msg.set_content(html_body, subtype="html")
        with smtplib.SMTP(self.host, self.port) as s:
            s.starttls()
            s.login(self.username, self.password)
            s.send_message(msg)


def get_sender() -> Sender:
    st = get_auth_settings()

    host = st.smtp_host
    user = st.smtp_username
    pw = st.smtp_password.get_secret_value() if st.smtp_password else None
    frm = st.smtp_from

    configured = all([host, user, pw, frm])

    # In prod, hard error if SMTP is not configured.
    if CURRENT_ENVIRONMENT == PROD_ENV and not configured:
        raise RuntimeError(
            "SMTP is required in prod for verification emails. Configure AUTH_SMTP_* envs."
        )

    # Dev fallback: console sender if not configured
    if not configured:
        return ConsoleSender()

    return SMTPSender(host, st.smtp_port, user, pw, frm)



from __future__ import annotations

import logging
from typing import Annotated, AsyncIterator, Tuple

from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)

from svc_infra.api.fastapi.deps import Require
from svc_infra.db.sql.utils import _coerce_to_async_url

logger = logging.getLogger(__name__)

_engine: AsyncEngine | None = None
_SessionLocal: async_sessionmaker[AsyncSession] | None = None


def _init_engine_and_session(
    url: str,
) -> Tuple[AsyncEngine, async_sessionmaker[AsyncSession]]:
    async_url = _coerce_to_async_url(url)
    if async_url != url:
        logger.info(
            "Coerced DB URL driver to async: %s -> %s",
            url.split("://", 1)[0],
            async_url.split("://", 1)[0],
        )
    engine = create_async_engine(async_url)
    session_local = async_sessionmaker(engine, expire_on_commit=False)
    return engine, session_local


def initialize_session(url: str) -> None:
    """Create engine + sessionmaker and store them in this module."""
    global _engine, _SessionLocal
    _engine, _SessionLocal = _init_engine_and_session(url)


async def dispose_session() -> None:
    """Dispose engine and clear globals."""
    global _engine, _SessionLocal
    if _engine is not None:
        await _engine.dispose()
    _engine = None
    _SessionLocal = None


async def get_session() -> AsyncIterator[AsyncSession]:
    if _SessionLocal is None:
        raise RuntimeError("Database not initialized. Call add_sql_db(app, ...) first.")
    async with _SessionLocal() as session:
        try:
            yield session
            await session.commit()
        except Exception:
            await session.rollback()
            raise


SqlSessionDep = Annotated[AsyncSession, Require(get_session)]

__all__ = ["SqlSessionDep", "initialize_session", "dispose_session"]



from __future__ import annotations

from pathlib import Path
from typing import Optional

import typer

from svc_infra.app.env import find_env_file, load_env_if_present
from svc_infra.app.root import resolve_project_root


def pre_cli(app: typer.Typer) -> None:
    @app.callback()
    def _bootstrap(
        env_file: Optional[Path] = typer.Option(
            None,
            "--env-file",
            dir_okay=False,
            exists=False,
            resolve_path=True,
            help="Path to .env (defaults to auto-discovery from CWD upward).",
        ),
        no_env: bool = typer.Option(False, "--no-env", help="Skip auto-loading .env"),
        override_env: bool = typer.Option(
            False, "--override-env", help="Allow .env to override existing vars"
        ),
    ):
        if no_env:
            return
        root = resolve_project_root()
        path = env_file or find_env_file(start=root)
        load_env_if_present(path, override=override_env)
